# Copyright (C) 2007 Wes Hardaker
# License: GNU GPLv2.  See the COPYING file for details.
package GeoDB;

use strict;

use GeoDB::Utils;
use GeoDB::DB;
use GeoDB::Set;
use GeoDB::Waypoint;
use GeoDB::Waypoint::Geocache;
use GeoDB::Waypoint::Waymark;
use GeoDB::Waypoint::Wigle;
use GeoDB::Waypoint::Geodining;
use Config;

use Text::Balanced qw(extract_bracketed extract_quotelike);

our $VERSION = "0.97";
our @ISA=qw(GeoDB::DB);

our %sqlxlat =
  ( '&&' => ' and ',
    '||' => ' or ',
    ')' => ')',
    '(' => '(',
    '' => '',
  );

# note: this new clause is used by most sub-modules too, altering it
# will alter them.
sub new {
    my $type = shift;
    my ($class) = ref($type) || $type;
    my $self = {};
    %$self = @_;
    bless($self, $class);
    $self->init_global($self->{'dbh'}) if ($self->{'dbh'});
    return $self;
}

######################################################################
# INITIALIZATION
#
sub init {
    my $self = shift;
    return $self->init_global(@_);
}

my $have_done_global_init = 0;
# initial needed setup by various modules
sub init_global {
    my $self = shift;

    return 0 if (!$self->{'dbh'});

    # run only once ever
    return 1 if ($have_done_global_init);
    $have_done_global_init = 1;

    # call modules that need it.

    # sub modules of waypoint MUST be called first to register themselves
    # so that the Waypoint can init based on those registrations.
    GeoDB::Waypoint::Geocache::init($self->{'dbh'});
    GeoDB::Waypoint::Waymark::init($self->{'dbh'});
    GeoDB::Waypoint::Wigle::init($self->{'dbh'});
    GeoDB::Waypoint::Geodining::init($self->{'dbh'});
    GeoDB::Waypoint::init($self->{'dbh'});
    return 1;
}

######################################################################
# ENGINE FUNCTIONS

#
# tries to require (load) and call new() on a generic perl module
#
sub new_engine {
    my ($self, $type, $parents, $superparent) = @_;
    my ($args, $argsep);

    $superparent = "GeoDB" if (!$superparent);
    $parents = "Search" if (!$parents);
    $parents = [$parents, "Everywhere"] if (ref($parents) ne 'ARRAY');

    # some things may pass :args terminated
    ($type, $argsep, $args) = ($type =~ /([\/\w]+)(:*)(.*)$/);

    # replace slashes with ::s
    $type =~ s/\//::/g;

    # capitalize first letter only of each module (enforced convention)
    $type =~ s/^(.)(.*)/uc($1) . lc($2)/e;
    $type =~ s/::(.)([^:]*)/"::" . uc($1) . lc($2)/ge;

    my $engine;
    my $err;
    for (my $i = 0; $i <= $#$parents && !$engine; $i++ ) {
	my $parent = $parents->[$i];
	GEODEBUG(6,
		 "Trying to load module '${superparent}::${parent}::$type'\n");

	# try and see if we can require the module
	# (if yes, it exists; if not skip it)
	my $res = eval "require ${superparent}::${parent}::$type;";
	if (!$res) {
	    my $potentialerr = $@;
	    if (!$err && $potentialerr !~ /Can.t locate/) {
		$err = "\n($potentialerr)";
	    }
	    next;
	}

	$engine = eval "new ${superparent}::${parent}::$type (args => \"$args\", type => \"$parents->[0]\")" ||
	  GEOERR("can't create a $parent engine for type='${superparent}::${parent}::$type';\n   (frequently syntax issue, no 'new' operator, or wrong package name ($@)\n");

	if (ref($engine) ne '') {
	    $engine->set_master($self);
	    if (!$argsep && !$self->get_option('nogui')) {
		$engine->prompt_args();
	    } else {
		$engine->parse_args($args);
	    }
	}

	return $engine;
    }
    GEOERR("Could not find $parents->[0] Engine Type '$type' $err\n")
      if (!$self->{'options'}{'allow-missing-engines'});
}

sub create_question_from_args {
    my ($self, $opt, $default) = @_;
    next if ($opt->{'nogui'});
    my $q =
      { type => 'text',
	name => $opt->{'name'},
	text => $opt->{'title'} || "$opt->{'name'}",
	helptext => $opt->{'help'},
	size => 60,
      };
    $q->{'default'} = $default if ($default);
    foreach my $passon (qw(type values labels indent helpdesc text doif
			   refresh_on_change default submit)) {
	if (exists($opt->{$passon})) {
	    $q->{$passon} = $opt->{$passon};
	    if ($q->{$passon} =~ /engine:(.*)/) {
		my $e = $self->new_engine($1, ["AutoGUI::Values"]);
		$q->{$passon} = $e->get_gui_values($q->{$passon});
	    }
	}
    }

    # special debugging
    $q->{'text'} = $opt->{'name'} . ":" . $q->{'text'}
      if ($self->{'main_options'}{'prepend_config_names'});

    # allow text override
    if ($self->{'options'}{'language'} && $self->{'langinfo'}) {
	foreach my $lookup (qw(text helpdesc help)) {
	    if (exists($self->{'langinfo'}{$self->{'options'}{'language'}}{$opt->{'name'}}{$lookup}) &&
		$self->{'langinfo'}{$self->{'options'}{'language'}}{$opt->{'name'}}{$lookup} ne '') {
		$q->{$lookup} =
		  $self->{'langinfo'}{$self->{'options'}{'language'}}{$opt->{'name'}}{$lookup};
	    }
	}
    }

    # special processing for most of the above (sigh)

    # values/labels become an array
    $q->{'values'} = [split(/,/,$q->{'values'})]
      if(exists($q->{'values'}) && ref($q->{'values'}) eq '');
    $q->{'labels'} = [split(/,/,$q->{'labels'})]
      if(exists($q->{'labels'}) && ref($q->{'labels'}) eq '');

    # doif is either a single word, or a word=value
    # (XXX: eventually allow arbitrary code refs somehow?)
    $q->{'doif'} = eval "sub { return qwparam($q->{'doif'}); }"
      if (exists($q->{'doif'}) && ref($q->{'doif'}) ne 'CODE' && 
	  $q->{'doif'} !~ /=/);
    if (exists($q->{'doif'}) && ref($q->{'doif'}) ne 'CODE' && 
	$q->{'doif'} =~ /=/) {
	my ($left, $right) = ($q->{'doif'} =~ /([^=]+)=(.*)/);
	$q->{'doif'} =
	  eval "sub { GEODEBUG(5,\"GUI doif checking: $left = $right \" . qwparam(\"$left\") . \":\\n\"); return (qwparam(\"$left\") eq \"$right\"); }";
    }

    # Auto-set the type to 'menu' if there is a labels or a value clause
    $q->{'type'} = 'menu'
      if ((exists($q->{'labels'}) || exists($q->{'values'})) &&
	  !exists($opt->{'type'}));
    return $q;
}

sub prompt_args {
    my ($self) = @_;
    my $qw = $self->get_qwizard();
    return if (!$qw);

    my $pluginname = ref($self);
    my $displayname = $pluginname;
    $displayname =~ s/^GeoDB:://;
    $displayname =~ s/::/ /;
    $displayname =~ s/::/\//g;
    my $options = eval "[\@${pluginname}::GEOOPTIONS];";
    my @options =@$options if ($options);

    return if ($#options == -1);

    my @qs;
    my $q;
    foreach my $opt (@options) {
	if (ref($opt) ne 'HASH') {
	    push @qs, { type => 'label', text => '', values => '' };
	    next;
	}
	next if ($opt->{'nogui'});
	$q = $self->create_question_from_args($opt);
	push @qs, $q;
    }

    $qw->merge_primaries(
    {
     'askqs' => 
     { title => "Parameters for the $displayname plugin",
       questions => \@qs,
       post_answers => 
       [[sub {
	    my $opts = $_[$#_-1];
	    my $self = $_[$#_];
	    foreach my $opt (@options) {
		next if (ref($opt) ne 'HASH');
		$self->{'options'}{$opt->{'name'}} = qwparam($opt->{'name'})
		  if (qwparam($opt->{'name'}) ne '');
	    }
	},\@options, $self]],
     },
    }
			);
    $qw->magic('askqs');
}

sub set_master {
    my ($self, $master) = @_;
    $self->{'master'} = $master;
}

sub get_master {
    my ($self) = @_;
    while (exists($self->{'master'})) {
	$self = $self->{'master'};
    }
    return $self;
}

sub get_option {
    my ($self, $opt) = @_;
    $self = $self->get_master();
    return $self->{'main_options'}{$opt} if (exists($self->{'main_options'}));
    return $main::opts{$opt};
}

sub set_option {
    my ($self, $opt, $value) = @_;
    $self = $self->get_master();
    if (exists($self->{'main_options'})) {
	$self->{'main_options'}{$opt} = $value;
    } else {
	$main::opts{$opt} = $value;
    }
}

sub save_option {
    my ($self, $opt, $value) = @_;
    $self = $self->get_master();
    return if ($self->get_option($opt) eq $value);
    $self->set_option($opt, $value);
    GEODEBUG(5, "saving option $opt, to $self->{'default_file'}\n");
    if ($self->{'default_file'}) {
	open(O,">>$self->{'default_file'}");
	print O "$opt $value\n";
	close(O);
    }
}


sub get_default {
    my ($self, $opt, $paramname) = @_;
    $self = $self->get_master();
    return $self->{'main_options'}{'screen_defaults'}{$paramname}{$opt}
      if (defined($paramname) &&
	  exists($self->{'main_options'}) &&
	  exists($self->{'main_options'}{'screen_defaults'}) &&
	  exists($self->{'main_options'}{'screen_defaults'}{$paramname}));
    return $self->{'main_options'}{'defaults'}{$opt}
      if (exists($self->{'main_options'}) &&
	  exists($self->{'main_options'}{'defaults'}));
    return $main::opts{'defaults'}{$opt};
}

sub save_default {
    my ($self, $opt, $value) = @_;
    $self = $self->get_master();
    return if ($self->get_default($opt) eq $value);
    $self->{'main_options'}{'defaults'}{$opt} = $value;
    if ($self->{'default_file'}) {
	open(O,">>$self->{'default_file'}");
	print O "default $opt $value\n";
	close(O);
    }
}

#
# parses specs of the form:
#   engtype1:arg1=arg2+engtype2:arg3=arg4:out+...
# returns an array of engines
#
sub get_engines_from_specs {
    my ($self, $specs, $parent, $superparent) = @_;
    my $args;
    my @engines;
    $specs = [$specs] if (ref($specs) ne 'ARRAY');

    foreach my $spec (@$specs) {
	while ($spec =~ s/^([^\+]+)(\+?)//) {
	    my $part = $1;
	    my $sep = $2;
	    GEODEBUG(3,"processing engine spec: $part => $spec\n");
	    my $e = $self->new_engine($part, $parent, $superparent);
	    if (ref($e) eq '') {
		$spec = $e . $sep . $spec;
	    } else {
		push @engines, $e;
	    }
	}
    }
    return \@engines;
}

#
# constructs a search "where" clause and arguments from a search spec string
#
sub construct_search {
    my ($self, $arg) = @_;

    my ($whereclause, $newwhereclause, $newargs);
    my @args;

    while ($arg) {
	if ($arg =~ s/^(\(|\))//) {
	    # allow (, or ) to pass through
	    $whereclause .= $1;
	} elsif ($arg =~ s/^(\&\&|\|\|)//) {
	    # allow && or || to pass through translated
	    $whereclause .= $sqlxlat{$1};
	} elsif ($arg =~ s/(\w+:|)([^\&\|\(\)]*)(\&\&|\|\||)//) {
	    # type:args and optionally &&, ||, ( or )
	    my ($type, $term, $clause) = ($1, $2, $3);

	    GEODEBUG(2,"Parsing search term: $type / $term / $clause\n");

	    GEODEBUG(5," type: $type\n");
	    my $engine = $self->new_engine("$type:$term");
	    if (ref($engine) eq '') {
		$arg = $engine . $arg;
		next;
	    }
	    ($newwhereclause, $newargs) = $engine->search($term, $type);
	    $whereclause .= " ($newwhereclause) $sqlxlat{$clause}";
	    GEODEBUG(3,"where clause part:      $newwhereclause\n");
	    GEODEBUG(3,"where clause arg count: $#$newargs\n");
	    push @args, @$newargs;
	} else {
	    print STDERR ("CAN'T PARSE ARGS:" . $arg);
	    exit;
	}
    }
    return ($whereclause, \@args);
}

# note: import is a reserved function name in perl, so we had to use import_it
sub import_it {
    my ($self, @files) = @_;
    my $set;
    foreach my $f (@files) {
	my $parenttype = "Import";
	GEODEBUG(1,"importing $f\n");
	while ($f) {
	    # +/; separated tokens $1+;$3
	    $f =~ s/([^\+;]+)([\+;]*)(.*)/$3/;
	    my ($it, $sep) = ($1, $2);
	    if (!$it) {
		$it = $f;
		$f = '';
	    }
	    if ($parenttype eq "Import") {
		my ($type, $file) = get_type_and_file($it);
		my $engine = $self->new_engine($type, $parenttype);
		if (ref($engine) !~ /GeoDB/) {
		    # must be a string replacement
		    $f = $engine . $sep . $f;
		    next;
		}
		$set->save($self->{'dbh'}) if ($set);
	        $set = $engine->import_it($file, $self->{'dbh'});
	    } else {
		my $engine = $self->new_engine($it, $parenttype);
		if (ref($engine) !~ /GeoDB/) {
		    # must be a string replacement
		    $f = $engine . $sep . $f;
		    next;
		}
		$engine->modify_set($set);
	    }
	    $parenttype = ((!$sep || $sep eq ';') ? "Import" : "Modify");
	}
	$set->save($self->{'dbh'}) if ($set);
	$set = undef;
    }
    return $set;
}

sub modify {
    my ($self, $set, $modifydescs) = @_;

    GEODEBUG(2,"modifying set as requested\n");

    my $engines = $self->get_engines_from_specs($modifydescs, "Modify");

    foreach my $engine (@$engines) {
	$engine->modify_set($set);
    }
}

#
# export data
#
sub export_it {
    my ($self, $exportdescs, $origset) = @_;
    $exportdescs = [$exportdescs] if (ref($exportdescs) ne 'ARRAY');

    my $set = $origset->clone();

    foreach my $f (@$exportdescs) {
	# modtype1:args+modtype2:args+outputfile.type
	GEODEBUG(5,"export spec: $f\n");
	while ($f) {
	    GEODEBUG(6,"export processing: $f\n");
	    # contains a modify part
	    my $sep;
	    ($sep) = ($f =~ /([\+;])/);
	    if ($sep && $sep eq '+') {
		$f =~ s/([^\+;]+)\+//;
		my $part = $1;
		my $engine = $self->new_engine($part, "Modify");
		if (ref($engine) eq '') {
		    # it returned a string; keep going with it
		    $f = $engine;
		} else {
		    # no other parts left, this is indeed the export engine
		    GEODEBUG(2,"modifying using " . ref($engine) . "\n");
		    $engine->modify_set($set);
		}

   	    # contains a single part, potentially the last export part
	    } else {
		$f =~ s/([^;]+);*//;
		my $part = $1;

		my ($type, $file);
		if ($part !~ /:/ || $part =~ /^[a-zA-Z]:[^:]*$/) {
		    GEODEBUG(3,"contains no : sections: $part\n");
		    ($type, $file) = get_type_and_file($part);
		    $part = "$type:$file";
		} else {
		    GEODEBUG(3,"assuming complete: $part\n");
		}

		my $engine = $self->new_engine($part, "Export");
		if (ref($engine) eq '') {
		    # it returned a string; keep going with it
		    $f = $engine . ($sep || "") . $f;
		} else {
		    # no other parts left, this is indeed the export engine
		    GEODEBUG(2,"exporting using " . ref($engine). "\n");
		    # pull out the args differently than the normal parsing
		    ($type, $file) = get_type_and_file($part) if (!$file);
		    my $engine = $self->new_engine($type, "Export");
		    $engine->export_it($file, $set);
		    $set = $origset->clone() if ($f);
		}
	    }
	}
    }
}

#
# take a search term (arg) and load it into (set or created) using options.
#   - returns either the passed set or a new one if one wasn't passed.
sub search {
    my ($self, $arg, $options) = @_;

    my $searchset = $options->{'searchset'} || GeoDB::Set->new();
    $searchset->set_master($self->get_master());

    my ($whereclause, $args) = $self->construct_search($arg);

    GEODEBUG(2,"final search clause: $whereclause\n");
    GEODEBUG(2,"final search args:   ", join(", ",@$args),"\n");

    my $ids = get_caches_from_where_clause($self->{'dbh'}, $whereclause, $args,
					   $options->{'limit'});
    if ($ids) {
	$searchset->add_by_ids(@$ids);
    } else {
	GEOERR("no waypoints found given the supplied search terms");
    }

    # save if they indicated we should
    $searchset->save_set($self->{'dbh'}, $options->{'name'})
      if ($options->{'name'} && $self->{'dbh'});

    return $searchset;
}

#######################################################################
# SCRIPTS                                                             #
#######################################################################

sub pull_gui_parts {
    my ($self, $buf) = @_;
    my @parts;
    foreach my $part (qw(title introduction)) {
	my ($extracted) = ($buf =~ /\%${part}\s*([^\%]+)/);
	$extracted ||= "";
	$extracted =~ s/^\s*//;
	$extracted =~ s/\s*$//;
	push @parts, $extracted;
    }
    return @parts;
}

our $globalgui;

sub process_includes {
    my ($self, $buf) = @_;

    while ($buf =~ s/%include\s*(.*) (.+)\n/$self->file_to_buf($2,"",$1)/e) {
    }
    return $buf;
}

sub run_script {
    my ($self, $buf, $args, $name) = @_;

    my $set = new GeoDB::Set;
    my $x;

    my $options_save = $self->{'options'};
    $self->{'options'} = undef;
    $self->parse_args($args);

    # convert from a file if needed
    if ($buf =~ s/^file:(.*)//) {
	my $fname = $1;
	$name = $fname if (!$name);
	$buf = $self->file_to_buf($fname)
    }

    $buf = $self->process_includes($buf);

    # pull the title and introduction out
    my ($title, $intro) = $self->pull_gui_parts($buf);

    # process %[] params
    $buf =  $self->process_string_params($buf, '[', ']',
					 ($self->get_option('nogui') ? 0 : 1),
					 $title, $intro, $name);
    ($buf) =  $self->deifify_text($buf);

    $globalgui = $self->get_option('nogui');
    $self->set_option('nogui', 1);

    GEODEBUG(1,"running a script\n");
    my $mode = '';
    my $text = '';
    my $scriptargs;
    my @scriptargs;
    my %scriptargs;

    while ($buf =~ s/^([^\n]*)\n//) {
	my $line = $1;
	next if ($line =~ /^\s*#/);
	next if ($line =~ /^\s*$/);
	if ($line =~ /^\%\s*(\w+)\s*(.*)/) {
	    my $newmode = $1;
	    my $newscriptargs = $2;
	    $self->run_mode_on_text($mode, $text, $set, $self->{'dbh'}, $args,
				    \%scriptargs);
	    $text = '';
	    $mode = $newmode;
	    $scriptargs = $newscriptargs;
	    @scriptargs = split(/ /,$scriptargs);
	    %scriptargs = map { split(/=/,$_); } @scriptargs;
	    next;
	} else {
	    if (!$scriptargs{'leavewhitespace'}) {
		$line =~ s/^\s*//;
		$line =~ s/\s*$//;
	    } else {
		$line .= "\n";
	    }
	    $line =~ s/^"""//;
	    $line =~ s/"""$//;
	    $text .= $line;
	}
    }
    $self->run_mode_on_text($mode, $text, $set, $self->{'dbh'}, $args,
			    \%scriptargs);
    $self->{'options'} = $options_save;
    $self->set_option('nogui', $globalgui);
}

#
# currently parses text only for %if/%else/%endif pairs.
#  - really not the right place to do this; should operate as working
#  - that would require a 'noop' mode for gui var extration...
sub deifify_text {
    my ($self, $buf) = @_;
    my $returntext = "";
    my ($addtext, $matchtext, $ignoretext);

    while ($buf =~ s/^([^\n]*\n)//) {
	my $line = $1;
	next if ($line =~ /^\s*#/);
	next if ($line =~ /^\s*$/);

	# look for marking section ( ^%KEYWORD )
	if ($line =~ /^\%\s*(endif|else|endforeach)/) {
	    return ($returntext, $buf, $1);
	} elsif ($line =~ /^\%\s*(\w+)\s*(.*)/) {
	    my $keyword = $1;
	    my $otherargs = $2;
	    if ($keyword eq 'if') {
		if ($otherargs && eval "$otherargs") {
		    # true if statement
		    ($addtext, $buf, $matchtext) =
		      $self->deifify_text($buf);
		    $returntext .= $addtext;
		    if ($matchtext eq 'else') {
			($ignoretext, $buf, $ignoretext) =
			  $self->deifify_text($buf);
		    }
		} else {
		    # false if statement
		    ($ignoretext, $buf, $matchtext) =
		      $self->deifify_text($buf);
		    if ($matchtext eq 'else') {
			($addtext, $buf, $matchtext) =
			  $self->deifify_text($buf);
			$returntext .= $addtext;
		    }
		}
	    } elsif ($keyword eq 'foreach') {
		my @args = split(/\s+/,$otherargs);
		($addtext, $buf, $matchtext) = $self->deifify_text($buf);
		my $paramname = shift @args;
		foreach my $arg (@args) {
		    my $newsectiontext = $addtext;
		    $newsectiontext =~ s/\%\{$paramname\}/$arg/g;
		    $returntext .= $newsectiontext;
		}
	    } else {
		$returntext .= $line;
	    }
	} else {
	    $returntext .= $line;
	}
    }
    return ($returntext, $buf, '');
}

sub replace_webfile {
    my ($self, $arg) = @_;
    my $webdir = $self->get_config_dir() . "/webfiles/";
    $arg =~ s/webfile:/$webdir/g;
    return $arg;
}

sub run_mode_on_text {
    my ($self, $mode, $text, $set, $dbh, $args, $scriptargs) = @_;

    $dbh = $self->{'dbh'} if (!$dbh);
    if ($scriptargs->{'allowgui'}) {
	$self->set_option('nogui', $globalgui);
    }
    if ($mode eq 'search') {
	GEODEBUG(2,"script: searching using $text\n");
	$set->clear(); # XXX: make optional
	$scriptargs->{'searchset'} = $set;
	$scriptargs->{'dbh'} = $dbh;
	$self->search($text, $scriptargs);
	GEODEBUG(3, "search results: " . $set->num() . "\n");
    } elsif ($mode eq 'export') {
	GEODEBUG(2,"script: exporting using $text\n");
	$self->export_it($text, $set);
    } elsif ($mode eq 'import') {
	GEODEBUG(2,"script: importing using $text\n");
	$self->import_it($text);
    } elsif ($mode eq 'saveset') {
	GEODEBUG(2,"saving set as name: $scriptargs->{'name'}\n");
	$set->save_set($dbh, $scriptargs->{'name'});
    } elsif ($mode eq 'modify') {
	GEODEBUG(2,"script: modifying using $text\n");
	$self->modify($set, $text);
    } elsif ($mode eq 'display') {
	GEODEBUG(2,"script: displaying using $text\n");
	$self->display($set, $text, $dbh);
    } elsif ($mode eq 'print') {
	$text = $self->process_string_params($text,'[',']');
	GEODEBUG(2,"script: printing text\n");
	print $text,"\n";
    } elsif ($mode eq 'perl') {
	$text = $self->process_string_params($text,'[',']')
	  unless ($scriptargs->{'dontexpand'});
	GEODEBUG(2,"evaluating embedded perl\n");
	my $result = eval $text;
	if ($@) {
	    print STDERR "dump:$text\n";
	    GEOERR("evaluated perl died: $@\n");
	}
    } elsif ($mode eq 'script') {
	my ($scriptname, $args) = ($text =~ /([^:]+):*(.*)/);
	GEODEBUG(2,"script: running script $scriptname\n");
	$self->run_script("file:" . $scriptname, $args, $scriptname);
    } elsif ($mode eq 'system') {
	$text = $self->replace_webfile($text);
	my (@args) = split(/ +/,$text);
	GEODEBUG(2,"script: running system command: $text\n");
	system(@args);
    } elsif ($mode eq 'copy') {
	GEODEBUG(2,"script: copying files\n");
	if ($text =~ /^([^,]+),([^,]+),(.*)/) {
	    my ($filename, $to, $fromURL) = ($1,$2,$3);
	    # filename,to,fromURL
	    my $dir = $self->get_config_dir() . "/webfiles/";
	    grab_url($fromURL, "$dir/$filename");
	    copy_file("$dir/$filename", $to);
	} else {
	    my @copies = split(/,/,$text);
	    copy_file($self->replace_webfile($copies[0]), $copies[1]);
	}
    } elsif ($mode eq 'webfile') {
	my ($file, $url) = ($text =~ /^([^,]+),(.*)/);
	GEODEBUG(2,"script: storing a URL $url\n");
	my $dir = $self->get_config_dir() . "/webfiles/";
	mkdir($dir) if (! -d $dir);
	grab_url($url, "$dir/$file")
    } elsif ($mode eq 'ignore' || $mode eq 'introduction' || $mode eq 'title') {
	GEODEBUG(2,"Ignoring text in the ignore section\n");
    } elsif ($mode eq 'save') {
	GEODEBUG(2,"saving set back to the DB");
	$set->load();  # ensure everything got loaded before we save it!
	               # XXX: have the save routines only save what got loaded
	$set->save();
    } elsif ($mode eq '') {
	GEOERR("no mode to run script text on: $text\n") if ($text ne '');
    } else {
	GEOERR("unknown script mode: $mode\n");
    }
    $self->set_option('nogui', 1);
}

sub find_file_from_paths {
    my ($file, $paths, %options) = @_;

    foreach my $path (@$paths) {
	if (-f "$path$file") {
	    return "$path$file";
	}
    }
    GEODEBUG(1,"paths searched: ", join(",",@$paths), "\n");
    GEOERR("Could not find a required file: '$file'") if ($options{'fatal'} == 1);
    return;
}

sub file_to_buf {
    my ($self, $file, $dirname, $args) = @_;
    $dirname ||= "scripts";
    my $buf;

    GEODEBUG(99, ("-" x 40) . $file . " to buf\n");

    # parsing args
    my %options;
    if (defined($args)) {
	my @args = split(/\s+/,$args);
	foreach my $arg (@args) {
	    my ($left, $right) = split(/=/,$arg);
	    $options{$left} = $right;
	}
    }


    my $specifiedpath = $self->get_option('scripts');
    my @searchpaths;
    if ($specifiedpath) {
	foreach my $path (split(/\s*,\s*/,$specifiedpath)) {
	    $path .= "/" if ($path !~ /\/$/);
	    push @searchpaths, $path;
	}
    } else {
	@searchpaths = ("",  # right here
			$self->get_config_dir() . "/$dirname/",
			"/usr/share/geoqo/scripts/",
			# windows
			$Config{'prefix'} . "/site/share/geoqo/scripts/",
		       );
    }

    my @files;

    # handle special types of includes
    if (exists($options{'type'}) && $options{'type'} ne "") {
	# search for a bunch of files, and include them all
	if ($options{'type'} eq 'hooks') {
	    GEODEBUG(7,"expanding hooks for $file\n");
	    # the actual file name is expected to be a 'glob' pattern
	    foreach my $path (@searchpaths) {
		GEODEBUG(8,"looking for $path$file\n");
		foreach my $globfile (glob("$path$file")) {
		    if (-f "$globfile") {
			GEODEBUG(9,"found one: $globfile\n");
			push @files, "$globfile";
		    }
		}
	    }
	} else {
	    GEOERR("illegal include type: $options{'type'}\n");
	}
    } else {
	# just a normal single file include
	GEODEBUG(7,"looking for single script: $file\n");
	push @files, find_file_from_paths($file, \@searchpaths, fatal => 1);
    }

    # we should find at least one
    GEOERR("can't find script: $file\n") if ($#files == -1);

    # sort based on alphabetical of last path component name
    @files = sort {
	my ($f1) = ($a =~ /([^\/]*)$/);
	my ($f2) = ($b =~ /([^\/]*)$/);
	return $f1 cmp $f2;
    } @files;

    my @titles;
    for (my $i = 0; $i <= $#files; $i++) {
	my $filebuf;
	my $num = $i+1;
	if (exists($options{'autoif'})) {
	    $filebuf .= "\%if \"\%[$options{'autoif'}]\" eq '$num'\n";
	}
	my $file = $files[$i];
	open(INFILE, $file);
	while (<INFILE>) {
	    s/\%{INCLUDENUMBER}/$num/g;
	    $filebuf .= $_;
	}
	close(INFILE);
	if (exists($options{'autoif'})) {
	    $filebuf .= "\%endif\n";
	}
	my ($title) = ($filebuf =~ /^#\s*scriptname:\s*([^\n]*)/m);
	$titles[$i] = $title || "unknown title #$i";
	$buf .= $filebuf;
    }

    if ($options{'hookheader'}) {
	my $hookfile = find_file_from_paths($options{'hookheader'},
					    \@searchpaths, fatal => 1);
	my $values;
	my $valuelabels;
	my $labels;

	for (my $i = 0; $i <= $#files; $i++) {
	    my ($title, $intro) = $self->pull_gui_parts($buf);
	    my $num = $i+1;
	    $values .= "$num,";
	    $valuelabels .= "$num,$titles[$i],";
	    $labels .= "$titles[$i],";
	}

	$values      =~ s/,$//;
	$valuelabels =~ s/,$//;
	$labels      =~ s/,$//;

	my $hookbuf;
	open(INFILE, $hookfile);
	while (<INFILE>) {
	    s/\%{VALUES}/$values/g;
	    s/\%{TITLES}/$labels/g;
	    s/\%{(VALUELABELS|LABELVALUES|LABELS)}/$valuelabels/g;
 	    next if (/^\s*#/);
 	    next if (/^\s*$/);
	    $hookbuf .= $_;
	}
	close(INFILE);
	$buf = $hookbuf . $buf;
    }

    GEODEBUG(99, ("-" x 40) . " file to buf: \n" .  $buf . ("-" x 40) . "\n");

    $buf =~ "\n" if ($buf !~ /\n$/);
    return $buf;
}

#
# loads display engines and calls them on a search set
#
sub display {
    my ($self, $searchset, $displist, $dbh) = @_;

    my @displist = split(/;/,$displist);

    # remove the output file
    if ($self->get_option('to-file')) {
	unlink($self->get_option('to-file'));
    }

    foreach my $disparg (@displist) {
	my $engine = $self->new_engine($disparg, "Display");
	my $text = $engine->display($searchset, $dbh);
	if ($text) {
	    # don't create a new one if DNE already
	    my $qw = $self->{'qwizard'};
	    if ($qw && !$self->get_option('toconsole')) {
		# display results in the existing window
		my %primaries =
		  ('geoqodisplay' =>
		    { title => 'GeoQO Output',
		    name => 'geoqodisplay',
		    questions => 
		    [{ type => 'paragraph',
		       preformatted => 1,
		       values => sub { return [$text]; }
		     }]
		    });
		$qw->merge_primaries(\%primaries);
		$qw->magic('geoqodisplay')
	    } elsif ($self->get_option('to-file')) {
		open(OOO,">>" . $self->get_option('to-file'));
		print OOO $text;
		close(OOO);
	    } else {
		# display results on the console, or whatever.
		print $text if ($text);
	    }
	}
    }
}


######################################################################
# GUI
#

sub get_qwizard {
    my ($self, $brandnew, $onlyifexists) = @_;

    $self = $self->get_master();
    return if($onlyifexists && !$self->{'qwizard'});

    eval "require QWizard;" || return;
    require QWizard;
    import QWizard;
    return $self->{'qwizard'}
      if (exists($self->{'qwizard'}) && !$brandnew);

    $self->{'qwizard'} = new QWizard(title => "GeoQO $VERSION");
    return $self->{'qwizard'};
}

######################################################################
# PARSING

#
# parses a string with a specified delimeter into pieces respecting
# paren matching.
#

# if delim=,
# abc,(def,[hik),xyz
# => ('abc', 'def,[hik', 'xyz')
sub split_pieces {
    my ($self, $delim, $string) = @_;
    my (@parts, $part);
    while ($string) {
	if ($string =~ /^([\(\[\{])/) {
	    my $paren = $1;
	    # contained a starting paren of some kind;
	    ($part, $string) = extract_bracketed($string, $paren);
	    $string =~ s/^$delim//;
	    $part =~ s/^.//;
	    $part =~ s/.$//;
	} else {
	    ($part, $string) = ($string =~ /^([^$delim]*)$delim?(.*)/);
	}
	push @parts, $part;
    }
    return @parts;
}

#
# parses a format of type:file or file.type and returns type,file
#
sub get_type_and_file {
    my ($f) = (@_);
    GEODEBUG(2,"examining file name $f\n");
    my ($type, $file, $args, $sep);

    # first determine if there is a prefix:REST syntax
    # requires TWO letters for import prefix to get around an issue on
    # windows with paths like C:/blah/
    ($type, $sep, $file) = ($f =~ /([a-zA-Z0-9][a-zA-Z0-9]+)(:)(.*)/);

    # if there is file specified or there isn't but there is a
    # separator, then assume
    if ($file || (!$file && $sep)) {
	GEODEBUG(3,"import component specified directly: $type\n");
	# yay!
	# but... there may be arguments, ie: prefix:arg1=val1:file.something
	# in which case, arg1 and val1 may contain :s as well...
	# note that:
        #   system:cmd=gpsbable -i geo -f %{file:loc} -o ...
        # is common and thus you can't assume a :file.xyz suffix
	# thus the legal file parts are a-z,A-Z,0-9,. and /
	$f = $file;
	my $legal_file_pieces = '[- _\.a-zA-Z0-9\/\\]+';
	($args, $file) = ($f =~ /(.*):([- _\.a-zA-Z0-9\/\\]+)$/);
	if ($args && ($args =~ s/^([a-zA-Z])$// || 
		      $args =~ s/:([a-zA-Z])$//)) {
	    # windows style file with extra colon (ugh)
	    $file = $1 . ":" . $file;
	}
	if ($file) {
	    # we found both arguments and a file path
	    # (nothing to do)
	} else {
	    # no actual sub-pieces
	    if ($f =~ /^[-_\.a-zA-Z0-9\/\\]+$/) {
		# assume just a file name is present
		$file = $f;
		$args = "";
	    } else {
		# A colon exists in it; assume not a file
		# just $type should be returned (like system: above)
		$args = $f;
		$file = "";
	    }
	}
	GEODEBUG(3,"broken pieces into type=$type args=$args file=$file\n");
    } else {
	# there is no type: prefix, so assume the entire thing is a path
	$file = $f;
	GEODEBUG(3,"determining file type for $file\n");
	($type) = ($file =~ /.*\.([a-zA-Z0-9]+)$/);
	GEODEBUG(4,"extension says '$type'\n");

	# special rule for directories with a trailing /
	if (!$type && $file =~ /\/$/) {
	    $type = "Dir";
	}
# simply unsafe and prevents file-type skipping for zips/dirs/etc
#	$type ||= 'gpx';	# default to gpx if nothing else
    }

    return if (!$type);

    # now change the type to 1 upper case letter and everything else lower.
    $type =~ s/^(.)(.*)/uc($1) . lc($2)/e;

    $type ||= "";
    $args ||= "";
    $file ||= "";
    GEODEBUG(3,"file type parsing results: $type, $args, $file\n");
    return ("$type:$args", $file) if ($args);
    return ($type, $file);
}

#
# parse name=value,name2=value2 specifications and returns a hash
#
#   Also sets optionsarray and options in the (maybe) passed self ref
sub parse_args {
    my $self;
    $self = shift if (ref($_[0]) ne '');
    my ($specs, $args, %parse_options) = @_;
    $specs ||= "";

    $args = {} if (!defined($args));
    my @argarr;

    # retrieve the list of OPTION specs from the child module (if it has any) 
    my @options = eval "@" . ref($self) . "::GEOOPTIONS;";
    my $numopts = $#options;
    my %info = eval "%" . ref($self) . "::GEOINFO;";

    # specs = arg=val,...
    my @components;
    my $part;
    while ($specs) {
	if ($specs =~ /^\(/) {
	    ($part, $specs) = extract_bracketed($specs, '(');
	    $part =~ s/^\(//;
	    $part =~ s/\)$//;
	} else {
	    ($part, $specs) = ($specs =~ /([^,]*),*(.*)/)
	}
	push @components, $part;
    }
    foreach my $c (@components) {
        if ($c =~ /([^=<>]+)(==|=-|=\+|=!|=|<=>|<>|<|>)(.*)/) {
	    $args->{$1} = $3;
	    push @argarr, [$1, $3, $2];
	} else {
	    # assume a pre-defined option spec
	    my $opt = shift @options;
	    next if (ref($opt) ne 'HASH');
	    GEOERR("unknown value passed to " . ref($self) . ": $c\n")
	      if (!$opt && $numopts > -1 && !$info{'allowextraargs'});
	    next if (!$opt);
	    $args->{$opt->{'name'}} = $c;
	    push @argarr, [$opt->{'name'}, $c, '='];
	}
    }

    # set the defaults for the remainder of unspecified options.  maybe.
    foreach my $option (@options) {
	next if (ref($option) ne 'HASH');
	if (!defined($args->{$option->{'name'}})) {
	    if (exists($option->{'isconfig'})) {
		$args->{$option->{'name'}} =
		  $self->get_config($option->{'name'})
		    if ($self->get_config($option->{'name'}));
	    } else {
		$args->{$option->{'name'}} =
		  $self->get_default($option->{'name'})
		    if ($self->get_default($option->{'name'}));
	    }
	    $args->{$option->{'name'}} ||= $option->{'default'};
	}
    }

    $self->{'options'} = $args if ($self);
    $self->{'optionsarray'} = \@argarr if ($self);
    return \@argarr if ($parse_options{'wantarrayref'});
    return $args;
}

# chomps newlines off of beginning and end
sub chomp_spaces {
    my $string = shift;
    $string =~ s/^\s*//;
    $string =~ s/\s*$//;
    return $string;
}

sub process_token_definition {
    # takens a token string and parses it into an array:
    # format:  name,default,token1=value1,token2=(item1,item2,...),...
    # only required field: name
    # returns [name, default, token1, value1, token2, [item1, item2]]

    my ($string) = @_;
    my @output;

    # pull out the name
    $string =~ s/^([^,]+),{0,1}//;
    push @output, chomp_spaces($1);
    return \@output if ($string eq '');

    # pull out the default value
    # xxx: allow for quoted spots for including ,s in the default
    $string =~ s/^([^,]*)(,{0,1})//;
    push @output, chomp_spaces($1);
    return \@output if ($string eq '' && $2 eq '');

    GEODEBUG(7,"string piece: $output[0], $output[1]...\n");

    # extract the remainder token=value pairs into pieces
    while ($string) {
	GEODEBUG(9,"remaining: $string...\n");
	$string =~ s/^\s*([^,=]+)(=|,)\s*//;
	my $tokenname = $1;
	GEOERR("Can't get token name for parameter in script ($output[0])\n")
	  if (!$tokenname);

	# ensure that there was an = sign
	# XXX: at some point allow for ,tokenname,... with implicit 1 value?
	#   (nah:  something=, does same thing almost)
	GEOERR("expecting an = sign but got a ,\n") if ($2 eq ',');

	push @output, $tokenname;
	$string =~ s/^\s*//;
	if ($string =~ /^\(/) {
	    # list context; construct a list
	    my ($list, $rest) = extract_bracketed($string, '()');
	    $list =~ s/^\(//;
	    $list =~ s/\)$//;
	    # XXX: allow for ,s in output checking for quotedness
	    push @output, [split(/\s*,\s*/,$list)];
	    # (maybe) chomp the next comma
	    $rest =~ s/^\s*,*\s*//;
	    $string = $rest;
	} elsif ($string =~ /^\{/) {
	    # quoted literal bracked context
	    my ($txt, $rest) = extract_bracketed($string, '{}');
	    $txt =~ s/^\{//;
	    $txt =~ s/\}$//;
	    push @output, $txt;
	    # (maybe) chomp the next comma
	    $rest =~ s/^\s*,*\s*//;
	    $string = $rest;
	} elsif ($string =~ /^[\"\']/) {
	    # quoted string (probably contains a ,)
	    my ($value, $rest) = extract_quotelike($string);
	    $value =~ s/^[\'\"]//;
	    $value =~ s/[\'\"]$//;
	    push @output, $value;
	    # (maybe) chomp the next comma
	    $rest =~ s/^\s*,*\s*//;
	    $string = $rest;
	} else {
	    my ($value);
	    $string =~ s/^([^,]*),*\s*//;
	    $value = $1;
	    # trim the whitespace off the value
	    $value =~ s/\s*$//;
	    push @output, $value;
	}
    }
    return \@output;
}

# replaces strings with %{name} tokens with option values of the same name
sub process_string_params {
    my ($self, $string, $left, $right, $usegui, $title, $introduction,
	$paramname) = @_;
    $left ||= "{";
    $right ||= "}";
    my $regexp = "\%\\$left([^\\$right\/]+)/?([^\\/\\$right]*)/?([^\\$right]*)\\$right";
    my $qw;
    $qw = $self->get_qwizard()
      if ($usegui && $#{$self->{'optionsarray'}} == -1);

    GEODEBUG(5,"pulling part script pieces\n");
    my $parts = get_macro_parts('%', $left, $right, $string);

    for (my $i = 1; $i <= $#$parts; $i += 2) {
	# process just the expansion sections (odd elements of the array)
	$parts->[$i] = process_token_definition($parts->[$i]);
    }

    if ($qw) {
	my @qs;
	my @options;
	my %doneit;
	my $outstring = "";

	for (my $i = 1; $i <= $#$parts; $i += 2) {
	    # process just the expansion sections (odd elements of the array)

	    my ($name, $default, @remaining) = @{$parts->[$i]};

	    next if (exists($doneit{$name}));
	    $doneit{$name} = 1;
	    push @options, $name;

	    my %guiopts;
	    $guiopts{'default'} = $default if (defined($default));
	    $guiopts{'name'} = $name;
	    $guiopts{'text'} = $name;

	    for (my $i = 0; $i <= $#remaining; $i += 2)  {
		$guiopts{$remaining[$i]} = $remaining[$i+1]
	    }

	    if (exists($guiopts{'isconfig'})) {
		$guiopts{'default'} =
		  $self->get_option($name) if ($self->get_option($name));
	    } else {
		$guiopts{'default'} =
		  $self->get_default($name, $paramname)
		    if ($self->get_default($name, $paramname));
	    }

	    # special handling for the button type; allow it to be
	    # displayed again.
	    delete $doneit{$name} if (($guiopts{'type'} eq 'button' ||
				       $guiopts{'type'} eq 'label') &&
				      !exists($guiopts{'singleonly'}));

	    next if ($guiopts{'nogui'});

	    my $q = $self->create_question_from_args(\%guiopts);
	    push @qs, $q;
	}
	$qw->merge_primaries(
			     {
			      'scriptparams' => 
			      { title => ($title || 'Script Parameters'),
				introduction => $introduction,
				questions => \@qs,
				post_answers => 
				[[sub {
				      my $opts = $_[$#_-1];
				      my $self = $_[$#_];
				      foreach my $opt (@options) {
					  $self->{'options'}{$opt} = 
					    qwparam($opt)
					      if (qwparam($opt) ne '');
				      }
				  },\@options, $self]],
			      },
			     });
	$qw->magic('scriptparams') if ($#qs > -1);
    }
    for (my $i = 1; $i <= $#$parts; $i += 2) {
	# process just the expansion sections (odd elements of the array)

	# XXX: doesn't pull in default sfrom config?
	# set the part to either the expanded name, or the default value

	if (defined($self->{'options'}{$parts->[$i][0]})) {
	    $parts->[$i] = $self->{'options'}{$parts->[$i][0]};
	} elsif ($self->get_default($parts->[$i][0])) {
	    # XXX: should handle isconfig like above; ack!
	    $parts->[$i] = $self->get_default($parts->[$i][0]);
	} else {
	    # use the defined default
	    $parts->[$i] = $parts->[$i][1];
	    $parts->[$i] = "" if (!defined($parts->[$i]));
	}
    }

    # finally, join everything back together.
    return join("",@$parts);
}

#######################################################################
# PROGRESS BAR support
#
my $have_progress;
sub setup_progress {
    my ($self, $count, $title) = (@_);
    $self = $self->get_master();
    if (!$self->{'no-progress'} && !$self->get_option('no-progress') &&
	$self->get_option('debug') && $self->get_option('debug') < 4) {
	# don't event attempt to load this if no-progress is set so
	# that long load errors aren't printed to the console (happens
	# in make test)

	my $qw = $self->get_qwizard(0,1);
	if ($qw && ref($qw->{'generator'}) =~ /Gtk2/) {
	    # XXX: bad QWizard: no API to detect this
	    $self->{'qprogressbar'} = 1;
	    $self->{'progresstitle'} = $title;
	    $self->{'progresscount'} = $count || 1;
	} else {
	    $have_progress = eval "require Term::ProgressBar";
	    if ($have_progress) {
		$self->{'progressbar'} =
		  Term::ProgressBar->new({count => $count || 1,
					  name => $title,
					  remove => 1,
					  ETA => 'linear'});
		$self->{'progressbar'}->minor(0);
		$self->{'progresstitle'} = $title;
		$self->{'progresscount'} = $count || 1;
	    }
	}
    }
}

sub set_progress {
    my ($self, $value, $newtitle) = @_;
    $self = $self->get_master();
    return 1000000 if ($self->{'no-progress'} ||
		       $self->get_option('no-progress'));
    if ($self->{'progressbar'}) {
	return $self->{'progressbar'}->update($value);
    } elsif ($self->{'qprogressbar'}) {
	$self->{'progresstitle'} = $newtitle if ($newtitle);
	my $title = $self->{'progresstitle'} .
	  "($value/$self->{'progresscount'})";
	$self->{'qwizard'}->set_progress($value / $self->{'progresscount'},
					 $title);
	return (($value * 100) % $self->{'progresscount'});
    }
    return 1000000;  # if they're checking return codes for the next
                     # update count, this will likely keep them away for longer.
}

sub stop_progress {
    my ($self) = @_;
    $self = $self->get_master();
    return if ($self->{'no-progress'} || $self->get_option('no-progress') || 
               !$self->{'progressbar'});
    $self->{'progressbar'}->update($self->{'progresscount'});
    $self->{'qprogressbar'} = 0;
}


######################################################################
# SUBCLASS HELPER FUNCTIONS
#

sub get_temp_dir {
    my $self = shift;
    return $self->{'options'}{'tmpdir'} || $self->get_option('tmpdir') ||
      $self->get_config_dir() . "/temp";
}

sub get_config_dir {
    my $self = shift;
    return $self->{'options'}{'configdir'} || $self->get_option('configdir') ||
      $ENV{'HOME'} . "/.geoqo";
}

1;

=head1 NAME

GeoDB - Geographical waypoint database manipulation

=head1 SYNOPSIS

  my $geodb = new GeoDB();
  $geodb->init_dbh('sqlite2','mysqlitedb','localhost',
                   'dbuser','dbpassword', '9999');

  setup_db(); # XXX: currently this exists only in geoqo and needs to be moved

  $geodb->update_db('sqlite2');  # update to current db schema.

  # import a bunch of waypoints from a gpx file
  $geodb->import_it('myfile.gpx');

  # Extremely complex searches are available:
  #   (this loads a 'waypoint' and 'cache' module and passes the
  #   associated search parameters to them)
  my $resultset = $geodb->search('waypoint:type=Geocache&&cache:terrain>2');

  # modify the data
  #   (in this case, it calculates a distance from GCABCD and stores
  #   it in the distance data node)
  $geodb->modify($resultset, 'distance:from=GCABCD');

  # tag the waypoints as "cool"
  $geodb->modify($resultset, 'tag:cool');

  # search by tag:
  my $newsearchset = $geodb->search('tag:cool');

  # display the results in a list:
  $geodb->display($newsearchset, 'list:');

  # display with particular list format
  $geodb->display($newsearchset,
                  'list:format=%{8:ident} %{6:subtype} %{40:urlname}');

  # export the results to a file:
  $geodb->export_it(['myextract.gpx', 'mylocfile.loc'], $newsearchset);

  # finally it has it's own scripting language as well:
  # (see the geoqo manual page for how to write scripts)
  $geodb->run_script('file:myscript','arg1=foo,arg2=42');

=head1 DESCRIPTION

The GeoDB perl module is the foundation for the geoqo application.  It
is designed to load, query, modify, search for, etc geographical
waypoints within an SQL database.  The GeoDB module is the very top
module of a hierarchy of supporting modules needed.  The GeoDB module
itself can serve as the primary interface to most of those other
modules, however.

For complete details on the geoqo application and this module, please
refer to the geoqo web site:

  http://www.geoqo.org/

=head1 MODULE INTERFACE

The GeoDB system is very extensible and new Display, Export, Import,
Modify, Search, and Waypoint types can be easily added to the system.
For details on writing plugins, see the documentation for
GeoDB::Display, ...

=head1 SEE ALSO

geoqo(1) -- this is a MUCH MUCH more extensive manual on the power of the system

=head1 AUTHOR

Wes Hardaker AKA Yamar <yamar@geoqo.org>

=head1 COPYRIGHT

GPLv2

=cut

