# Copyright (C) 2007 Wes Hardaker
# License: GNU GPLv2.  See the COPYING file for details.
package GeoDB::Import::Average;

require GeoDB::Import;
use GeoDB::Utils;
use GeoDB::Set;
use strict;
use QWizard;

our $VERSION = "0.97";

our @ISA = qw(GeoDB::Import);
our @GEOOPTIONS = 
({ name => 'ident',
   text => 'Identifier for new Waypoint:',
 },
{ name => 'type',
   text => 'Waypoint Type',
   type => 'menu',
   values => sub {
       return ['Waypoint', keys(%GeoDB::Waypoint::registered_types)];
   },
   default => 'Waypoint',
 },
 { name => 'from1',
   text => 'From Waypoint #1:',
 },
 { name => 'from2',
   text => 'From Waypoint #2:',
   helpdesc => '(Any number can be averaged; just keep listing them)',
 },
 { name => 'setname',
   text => 'Set Name for new Point:',
   default => 'created',
 },
 );

sub import_set {
    my ($self, $file, $dbh, %opts) = @_;

    GEODEBUG(1,
	     "creating new waypoint $self->{'options'}{'ident'} by averaging\n");
    my $waypoint =
      new GeoDB::Waypoint
	({
	  data =>
	  {
	   ident => $self->{'options'}{'ident'},
	  }
	 }
	);

    GEOERR("Failed to create waypoint: $self->{'options'}{'ident'}")
      if (!$waypoint);

    # loop through all specified waypoints and add their lat/lon
    # together to be averaged later.
    my ($n, $w);
    my $i = 1;
    while (exists($self->{'options'}{'from' . $i})) {
	my $from = 
	  new GeoDB::Waypoint({ data => 
				{'ident' => $self->{'options'}{'from' . $i}}
			      });
	$from->load_parts();
	GEODEBUG(3,"averaging in: " . $self->{'options'}{'from' . $i} . ": " .
		 $from->{'data'}{'lat'} . " / " . $from->{'data'}{'lon'} .
		 "\n");

	GEOERR("Failed to load average waypoint #$i: " .
	       $self->{'options'}{'from' . $i} . "\n")
	  if (!$from->{'data'}{'lat'});
	
	$n += $from->{'data'}{'lat'};
	$w += $from->{'data'}{'lon'};
	$waypoint->{'data'}{'type'} = $from->{'data'}{'type'}
	  if (!exists($waypoint->{'data'}{'type'}));
	$i++;
    }
    GEOERR("No from waypoints specified") if ($i == 1);

    GEODEBUG(2,"averaged " . ($i-1) ." waypoints\n");

    # set lat/lon based on the average math
    $waypoint->{'data'}{'lat'} = $n/($i-1);
    $waypoint->{'data'}{'lon'} = $w/($i-1);

    $waypoint->morph_to_specific();

    # put the results in a set
    my $set = new GeoDB::Set;
    $set->add($waypoint);
    $set->set_name($self->{'options'}{'setname'})
      if($self->{'options'}{'setname'});
    return $set;
}


1;

=head1 NAME

Average - Average a slew of waypoints to create a new one

=head1 DESCRIPTION

START HELP

  > geoqo -i average:ident=newwaypoint,from1=oldwaypoint1,from2=oldwaypoint2

The I<average> import plugin lets you create a new waypoint by
averaging the latitude and longitude of a number of other waypoints.
Useful to find the center of a number of waypoints, which is required
for a number of geocaching puzzle caches.

END HELP

=head1 AUTHOR

Wes Hardaker (AKA Yamar) <hardaker ATAT users.sourceforge.net> 

=cut
