# Copyright (C) 2007 Wes Hardaker
# License: GNU GPLv2.  See the COPYING file for details.
package GeoDB::Modify::Rate;

use strict;
use GeoDB::Modify;
use GeoDB::Utils;

our @ISA = qw(GeoDB::Modify);
our $VERSION = "0.97";

our @GEOOPTIONS =
  (
   { name => 'overall',
     type => 'menu',
     values => [1..10],
     title => 'Rating:',
     helpdesc => '10 = good',
   },
   { name => 'ratings',
     title => 'Ratings to apply:',
     helpdesc => '(format example: location=9/fun=3)',
   },
   { name => 'publish',
     type => 'checkbox',
     title => 'Publish ratings to rating server:',
   },
   { name => 'publishtags',
     indent => 1,
     type => 'checkbox',
     title => 'Publish only certain ratings:',
     helpdesc => '(slash "/" delimited)'
   },
   { name => 'ignoreratings',
     indent => 1,
     type => 'checkbox',
     title => 'Don\'t publish these ratings:',
     helpdesc => '(slash "/" delimited)',
   },
   { name => 'deleteratings',
     title => 'Delete these ratings:',
     helpdesc => '(slash "/" delimited)',
   },
   { name => 'fetch',
     type => 'checkbox',
     title => 'Fetch publicized ratings from rating server:',
   },
   { name => 'geoqouser',
     title => 'GeoQO Tag Server User Name:',
   },
   { name => 'geoqopass',
     title => 'GeoQO Tag Server Password:',
     type => 'hidetext',
   },
  );

sub set_up {
    my ($self, $set) = @_;
    if ($self->{'options'}{'ratings'}) {
	$self->{'ratings'} = [split(/\//,$self->{'options'}{'ratings'})];
	GEODEBUG(3,"doing rating list: " .
		 join(",",@{$self->{'ratings'}}) . "\n");
	# convert each element to a sub-array with name/value
	GEODUMP( $self->{'options'});
	GEODUMP( $self->{'ratings'});
	$self->{'ratinglist'} = [map {[split(/=/,$_)]} @{$self->{'ratings'}}];
	GEODUMP( $self->{'ratinglist'});
    }
    $self->{'deleteratings'}
      = [split(/\//,$self->{'options'}{'deleteratings'})];

    my %tagpublish;
    if ($self->{'options'}{'publishratings'}) {
	my @taglist = split("/",$self->{'options'}{'publishratings'});
	%tagpublish = map { $_, $_ } @taglist;
	$self->{'tagpublish'} = \%tagpublish;
    }

    my %ignoreratings;
    if ($self->{'options'}{'ignoreratings'}) {
	my @taglist = split("/",$self->{'options'}{'ignoreratings'});
	%ignoreratings = map { $_, $_ } @taglist;
	$self->{'ignoreratings'} = \%ignoreratings;
    }

}

sub modify {
    my ($self, $cache) = @_;

    $cache->load_parts({ratings => 1});
    if ($self->{'options'}{'deleteratings'}) {
	GEODEBUG(5, "deleting ratings: $self->{'options'}{'deleteratings'}\n");
	foreach my $rating (@{$self->{'deleteratings'}}) {
	    next if (($self->{'options'}{'publishratings'} &&
		      (!exists($self->{'ratingpublish'}{$rating}))) ||
		     ($self->{'options'}{'ignoreratings'} &&
		      exists($self->{'ignoreratings'}{$rating})));
	    $cache->remove_rating($rating);
	    GEODEBUG(6,"removing rating: '$rating' on $cache->{'data'}{'ident'}\n");
	}
    } else {
	if ($self->{'options'}{'overall'}) {
	    $cache->set_rating('overall', $self->{'options'}{'overall'});
	    GEODEBUG(6,"setting rating: 'overall' = $self->{'options'}{'overall'} on $cache->{'data'}{'ident'}\n");
	}
	if (exists($self->{'ratinglist'})) {
	    foreach my $rating (@{$self->{'ratinglist'}}) {
		$cache->set_rating($rating->[0], $rating->[1]);
		GEODEBUG(6,"setting rating: '$rating->[0]' = $rating->[1] on $cache->{'data'}{'ident'}\n");
	    }
	}
    }
    if ($self->{'options'}{'publish'}) {
	my $publishstr = "$cache->{data}{ident}";
	my $count = 0;
	foreach my $rating (keys(%{$cache->{'ratings'}})) {
	    next if ($rating =~ /predicted:/); # XXX: make optonal
	    next if (($self->{'options'}{'publishratings'} &&
		      !exists($self->{'ratingpublish'}{$rating})) ||
		     ($self->{'options'}{'ignoreratings'} &&
		      exists($self->{'ignoreratings'}{$rating})));
	    if ($cache->{'ratings'}{$rating}{'needspublished'}) {
		if (exists($cache->{'ratings'}{$rating}{'value'})) {
		    $publishstr .= ":$rating!$cache->{'ratings'}{$rating}{'value'}";
		} else {
		    $publishstr .= ":-$rating";
		}
	    }
	}
	$self->{'publishstr'} .= "/$publishstr"
	  if ($publishstr ne $cache->{'data'}{'ident'});
    }
    if ($self->{'options'}{'fetch'}) {
	$self->{'fetchstr'} .= "/$cache->{'data'}{'ident'}";
    }
}

sub clean_up {
    my ($self, $set) = @_;
    # XXX: write publish stuff here

    if ($self->{'options'}{'publish'} && $self->{'publishstr'}) {
	my $user = $self->{'options'}{'geoqouser'};
	my $pass = $self->{'options'}{'geoqopass'};
	my $server = $self->{'options'}{'geoqoupserver'} ||
	  'ratingsup1.geoqo.org';

	GEOERR("either geoqouser ($user) or geoqopass ($pass) or both wasn't set")
	  if (!$user || !$pass);

	# publish to the geoqo server
	maybe_load('LWP::UserAgent',
		   'The LWP::UserAgent perl module is required for uploading/downloading of ratings');
	my $ua = LWP::UserAgent->new(env_proxy => 1);
	my $req = HTTP::Request->new(POST => "http://$server/");
	$req->content_type('application/x-www-form-urlencoded');

	# create the content string
	# user=USER&pass=PASS&set=cache1:tag1:tag2/cache2:tag2:tag3

	$self->{'publishstr'} =~ s/^\///;
	$req->content("user=$user&pass=$pass&rate=$self->{'publishstr'}");
	my $res = $ua->request($req);

	GEODEBUG(1,"sent requested ratings to the rating server $server.  Status=" . $res->status_line . "\n");
	GEODEBUG(9,"sent rating string: $self->{'publishstr'}\n");
    }

    if ($self->{'options'}{'fetch'} && $self->{'fetchstr'}) {
	my $user = $self->{'options'}{'geoqouser'};
	my $pass = $self->{'options'}{'geoqopass'};
	my $server = $self->{'options'}{'geoqoupserver'} ||
	  'ratingsdown1.geoqo.org';

	GEOERR("either geoqouser ($user) or geoqopass ($pass) or both wasn't set")
	  if (!$user || !$pass);

	# publish to the geoqo server
	maybe_load('LWP::UserAgent',
		   'The LWP::UserAgent perl module is required for uploading/downloading of ratings');
	my $ua = LWP::UserAgent->new(env_proxy => 1);
	my $req = HTTP::Request->new(POST => "http://$server/");
	$req->content_type('application/x-www-form-urlencoded');

	$self->{'fetchstr'} =~ s/^\///;
	$req->content("user=$user&pass=$pass&getrates=$self->{'fetchstr'}");
	my $res = $ua->request($req);

	require XML::Simple;
	import XML::Simple qw(XMLin);
	my $ratings =
	  XMLin($res->content, ForceArray => [qw(rating ratings)]);

	$ratings = $ratings->{'ratings'};

	# update our local DB
	my $dbh = $self->get_master()->{'dbh'};
	GEOERR("can't find local database handle\n") if (!$dbh);

	my $insh = $dbh->prepare("insert into ratings(ident,ratingname,othersrating,needspublished,count,datefetched) values(?,?,?,0,?,date())");
	my $uph = $dbh->prepare("update ratings set count = ?, othersrating = ? where ident = ? and ratingname = ?");
	
	# process each received rating
	foreach my $rating (@$ratings) {
	    my $cache = $set->find_by_ident($rating->{'ident'});
	    foreach my $ratingname (keys(%{$rating->{'rating'}})) {

		# update the local structure as well as the DB
		if (exists($cache->{'ratings'}{$ratingname})) {
		    $cache->{'ratings'}{$ratingname}{'count'} = 
		      $rating->{'rating'}{$ratingname}{'count'};
		    $cache->{'ratings'}{$ratingname}{'othersrating'} = 
		      $rating->{'rating'}{$ratingname}{'value'};
		    $uph->execute($rating->{'rating'}{$ratingname}{'count'},
				  $rating->{'rating'}{$ratingname}{'value'},
				  $rating->{'ident'}, $ratingname);
		} else {
		    $cache->{'ratings'}{$ratingname} = 
		      {
		       count => $rating->{'rating'}{$ratingname}{'count'},
		       value => $rating->{'rating'}{$ratingname}{'value'},
		       needspublished => 0
		      };
		    $insh->execute($rating->{'ident'}, $ratingname,
				   $rating->{'rating'}{$ratingname}{'value'},
				   $rating->{'rating'}{$ratingname}{'count'});
		}
	    }
	}

	GEODEBUG(1,"fetched ratings from the rating server.  Status=" . $res->status_line . "\n");
    }
}


1;

=head1 NAME

Rate - Add ratings to caches

=head1 DESCRIPTION

START HELP

The "Rate" plugin allows you to add "ratings"s to your caches.

Example:

  geoqo -s cache:owner_name==Yamar -m 'rate:overall=4'

(at the moment "overall" is the only rating you can provide; more coming)

=head3 Publishing ratings

Your ratings may be published to a geoqo rating server, and you may pull
information about how other people have rated waypoints from the rating
server.  To do this you must first create an account on the rating server:

  http://ratingserver.geoqo.org/

Then put your created username and password into your
$HOME/.geoqo/config file like:

  geoqouser YOURUSERNAME
  geoqopass YOURPASSWORD

Then you can publish your ratings for a given set using:

  geoqo -s SEARCHTERMS -m rate:publish=1

Additionally you can fetch data about other people's ratings using:

  geoqo -s SEARCHTERMS -m rate:fetch=1

When you use the -S option to show a the data for a waypoint you will
see a list of all your ratings as well as a list of other peoples ratings
that you have fetched.

END HELP

=cut
