# Copyright (C) 2007 Wes Hardaker
# License: GNU GPLv2.  See the COPYING file for details.
package GeoDB::Modify::Translate;

use strict;
use GeoDB::Modify;
use GeoDB::Utils;

our @ISA = qw(GeoDB::Modify);

our $VERSION = "0.97";

maybe_load('WWW::Babelfish',"The WWW::Babelfish perl module is required to do cache translations.\n");

our @GEOOPTIONS = 
  (
   { name => 'source',
     help => 'The source language to translate from.',
     default => 'Spanish',
   },
   { name => 'destination',
     help => 'The destination language to translate to.',
     default => 'English',
   },
   { name => 'fields',
     help => 'A / separated list of fields to translate.',
     default => 'urlname',
   },
   { name => 'service',
     help => 'The name of the translation service to use.',
     default => 'Babelfish',
   }
  );

sub modify {
    my ($self, $cache) = @_;

    $cache->load_parts();
    my $xlator = new
      WWW::Babelfish(
		     service => $self->{'options'}{'service'} || 'Babelfish'
		    );

    my @fields = split(/\//,$self->{'options'}{'fields'} || "urlname");
    foreach my $f (@fields) {
	$cache->{'data'}{$f} = $xlator->translate(source => $self->{'options'}{'source'} || 'English',
						  destination => $self->{'options'}{'destination'} || 'French',
						  text => $cache->{'data'}{$f});
	if (!$cache->{'data'}{$f}) {
	    die("translation error: " . $xlator->error . "\n");
	}
    }
}

1;

=head1 NAME

Translate - Translate caches from one language to anthoer

=head1 DESCRIPTION

START HELP

By far, this is one of the most powerful modification plugins.  It
uses online services (like Google) to translate cache fields from one
language to another.  As an example, here are some of my caches:

  > geoqo -s cache:owner_name==Yamar&&waypoint:ident=GCJ -l
  num GCID     Name
  1   GCJYPW   Hoppin' mad about Davis
  2   GCJGRQ   Beneficial cache
  3   GCJKMD   Historical water tower
  4   GCJ7KT   Got Coffee?
  5   GCJ66H   Gone Fishin' (Yamar's unusual #4)
  6   GCJ491   O where art thou (Yamar's unusual #2)
  7   GCJ5DV   Inverted Cache (Yamar's unusual #3)

And here is what they look like when translated from English to Spanish:

  > geoqo -s cache:owner_name==Yamar&&waypoint:ident=GCJ -l -m translate:source=English,destination=Spanish
num GCID     Name
  1   GCJYPW   Hoppin' enojado sobre Davis
  2   GCJGRQ   Escondrijo beneficioso
  3   GCJKMD   Torre historica del agua
  4   GCJ7KT   Cafe Conseguido?
  5   GCJ66H   Fishin ido ' (4) inusual de Yamar #
  6   GCJ491   O donde mil del arte (2) inusual de Yamar #
  7   GCJ5DV   Escondrijo invertido (3) inusual de Yamar #

Valid languages to pick from:

  English
  Spanish
  French
  German
  Italian
  Korean
  Japanese
  Portuguese

=head3 Requirements

1) you must be connected to the net

2) you must have the WWW::Babelfish perl module installed.

END HELP

=head3 Options

=over

=item source=LANGUAGE

=item destination=LANGUAGE

Sets the FROM and TO languages.

=item service=SERVICE

There are multiple services to pick from that can perform translation:

  Babelfish   (the default)
  Google
  Yahoo

=item fields=FIELD1/FIELD2

A / separated list of fields to translate.  This defaults to just the
urlname field, which is what is shown in the -l or -d list output.

=back

END HELP

=cut
