# Copyright (C) 2007 Wes Hardaker
# License: GNU GPLv2.  See the COPYING file for details.
package GeoDB::Search::Track;

use strict;
use GeoDB::Search;
use GeoDB::Utils;
use XML::Simple;
use Date::Parse;
use Data::Dumper;

our @ISA = qw(GeoDB::Search);

our $VERSION = "0.97";

our @GEOOPTIONS =
  (
   { name => 'trackfile',
     help => 'A GPX trackfile to load and search by'
   },
   { name => 'speed',
     help => 'Only look for caches where your speed dropped below this',
     default => '3'
   },
   { name => 'delta',
     help => 'lat/lon delta to use looking for caches near a stopping point',
     default => '0.002'
   },
   { name => 'laterthan',
     help => 'Only use track points which are beyond a specifed time.',
   }
  );

sub search {
    my ($self, $criteria, $type) = @_;

    # load the track
    my $xdoc = XMLin($self->{'options'}{'trackfile'},
		     ForceArray => [qw(trkpt)]);
    if (exists($xdoc->{'trk'}{'trkseg'})) {
	# not in multiple track format...  make it...
	$xdoc->{'trk'} = { $xdoc->{'trk'}{'name'} => $xdoc->{'trk'} };
    }

    GEODEBUG(3, "searching for waypoints near a track\n");

    my $whereclause;
    my $delta = $self->{'options'}{'delta'};
    my %noredo;

    my $laterthan = $self->{'options'}{'laterthan'};
    if ($laterthan) {
	$laterthan = str2time($laterthan);
	if (!$laterthan) {
	    GEOERR("failed to parse laterthan option date/time value: ",
		   $self->{'options'}{'laterthan'});
	}
    }

    my ($llat, $llon);
    foreach my $track (keys(%{$xdoc->{'trk'}})) {
	my $count = 0;
	my $lasttime = 0;
	foreach my $pt (@{$xdoc->{'trk'}{$track}{'trkseg'}{'trkpt'}}) {
	    $count++;
	    my $thistime = str2time($pt->{'time'});
	    if ($laterthan && $thistime < $laterthan) {
		next;
	    }
	    my $distlast = calc_distance($pt->{'lat'}, $pt->{'lon'},
					 $llat, $llon);
	    my $speed = $distlast/($thistime - $lasttime)*60*60;
	    if ($lasttime && $speed < $self->{'options'}{'speed'}) {
		$whereclause .= " or (waypoints.lat > $pt->{'lat'} - $delta and waypoints.lat < $pt->{'lat'} + $delta and waypoints.lon > $pt->{'lon'} - $delta and waypoints.lon < $pt->{'lon'} + $delta)\n"
		  if (!exists($noredo{$llat}{$llon}));
		$noredo{$llat}{$llon} = 1;
		my $dist = calc_distance($pt->{'lat'} - $delta,
					 $pt->{'lon'} - $delta,
					 $pt->{'lat'} - $delta, $pt->{'lon'})
		  * 5280;
		$dist =~ s/^(......).*/$1/;

	    }
	    $llat = $pt->{'lat'};
	    $llon = $pt->{'lon'};
	    $lasttime = $thistime;
	}
    }

    $whereclause =~ s/^ or //;
    $whereclause = "( $whereclause )";
    return ($whereclause, []);
}


1;

=head1 NAME

Track - Searches for caches by using GPS track data

=head1 DESCRIPTION

START HELP

  > geoqo -s track:mytrackfile.gpx -d count:
  Search/Set Count: 3

The I<track> plugin loads track data from a GPX track file (which you
can retrieve from you gps using gpsbabel) and then compares the list
of caches in you database to this data.  It then matches all caches
where your speed dropped below a configurable speed (3MPH is the
default) and where the distance from the track is less than a
configurable amount (defaults to near 500 feet).

This is even more useful when combined with the Track modify plugin,
which can order caches based on the track and report the times you
found them as well.

If you know all the caches are in a given set, however, you should
probably limit the data it has to search through to a subset of you
total database if it runs slowly.

To get track data from the gps, use gpsbabel:

  gpsbabel -t -i garmin -f usb: -o gpx -F track.gpx

See the Modify/Track plugin for a detailed example

END HELP

=head1 AUTHOR

Wes Hardaker (AKA Yamar) <hardaker ATAT users.sourceforge.net>

=cut

