# Copyright (C) 2007 Wes Hardaker
# License: GNU GPLv2.  See the COPYING file for details.
package GeoDB::Search::Waypoint;

use strict;
use GeoDB::Search;
use GeoDB::Utils;
use GeoDB::Waypoint;

our @ISA = qw(GeoDB::Search);

our $VERSION = "0.97";

sub search {
     my ($self, $criteria, $type) = @_;

     GEODEBUG(3, "passing criteria: $criteria\n");
     return $self->get_search_tokens( $criteria, $type,
				      get_registered_cols_and_map() );
}


1;

=head1 NAME

Waypoint - query the waypoint list for waypoint attributes

=head2 DESCRIPTION

START HELP

  > geoqo -s waypoint:name=unusual -d count
  Search/Set Count: 6

The I<waypoint> search type is for searching for primary fields in a
waypoint record.  Possible search terms include B<anything> in the
type-specific searches (since waypoint: searches include
everything) and the following fields which apply to all waypoints
regardless of specific type:

	desc
	lat
	lon
	sym
	time
	type
	subtype
	url
	urlname
	cmt

The following are some examples of how to use the I<waypoint:> search
type.  Note that I<cache:> and I<waymark:> searches work the exact same
way and also do everything below, but limit your search to just
waypoints of those types.

The format for using the waypoint (or similarily for cache: or
waymark:) search terms is:

  geqo -s waypoint:I<TERMS>

where TERMS are a comma separated list of search terms.  Normally
these would be specified like:

  waypoint:FIELDNAME=VALUE

The I<FIELDNAME> indicates the field name you want to search (see
below), and the I<VALUE> is the value you want to search for.  The
value field is actually surrounded by wildcards, and thus searching
for "me" will match any value with "me" somewhere in it including "He
and I went to the store" and "sometimes".  Using two equals signs
turns off the automatic wildcard support (ie, name==me means search
for a name that is B<just> "me").

=begin comment

XXX: we should do this (again) at some point

Example of a generic search on any field (note, this is not as fast as
a more specific search so if you know what you want to search for,
don't use this generic form if possible):

Searches (-s) for "tubular" in any waypoint field and summarize (-d list)):

      > geoqo -s waypoint:tubular -d list
      num GCID     Name
        1 GCDFE0   Hidden Headstones
        2 GCE29A   Celebrated Cemetery
        3 GCHW9M   The eyes have it (Yamar's unusual #1)
        4 GCJ491   O where art thou (Yamar's unusual #2)
        5 GCJ5DV   Inverted Cache (Yamar's unusual #3)
        6 GCJ66H   Gone Fishin' (Yamar's unusual #4)
        7 GCQEBQ   Yamar's Unusual #6: The ORB
        8 GCB284   Don't Be A Drip!

=end comment

An example: find all waypoints with a urlname field containing "unusual":

      > geoqo -s waypoint:urlname=tubular -d list
      num GCID     Name
        1 GCDFE0   Hidden Headstones
        2 GCE29A   Celebrated Cemetery
        3 GCHW9M   The eyes have it (Yamar's unusual #1)
        4 GCJ491   O where art thou (Yamar's unusual #2)
        5 GCJ5DV   Inverted Cache (Yamar's unusual #3)
        6 GCJ66H   Gone Fishin' (Yamar's unusual #4)
        7 GCQEBQ   Yamar's Unusual #6: The ORB
        8 GCB284   Don't Be A Drip!

An example: find all geocaches owned by 'Spelunk':

      > geoqo -s cache:owner_name=spelunk -d list
      num GCID     Name
        1 GCJVTR   Here's to Your Health Cache
        2 GCJY4V   Sumac Grove Cache
        3 GCJZ2N   Wildlife View Cache
        4 GCKNZB   Misplaced Seed Cache
        5 GCKQBN   Out-N-About Cache
        6 GCKVFW   We Give-A-Hoot Cache
        7 GCNWM9   BUG JAR
        8 GCP5YP   Greenbelt  Hideaway
        9 GCPZK3   Davis Shopping Spree
       10 GCQ6Q8   BUG JAR II
       11 GCQ6TV   Great Experience Cache
       12 GCQ87Z   Here's to the Greenbelts
       13 GCQFYJ   Toby & Willie's Friends
       14 GCQH1M   Connect the Dots Cache
       15 GCTCC5   Canal Cache
       16 GCTFJH   The Micro Equalizer Cache
       17 GCVVWY   Spelunk's 'Fishy' Doings
       18 GCW26F   Mountain or Desert? Cache

Note that terms will match anything beyond the underscore ("_").  IE,
if you do a search for:

  waypoint:short=some text

It will search the short description for "some text" even though the
actual field name it'll search is "groundspeak_short_description".
Using an exact fully spelled out field name is the most precise, but
you can truncate names for convience.

Finally, you can use '*'s to indicate wildcard spots (I<u*sal> would
match "unusual" and "unit of sale").  By default geoqo already adds a
* to both the front and the back of your search time (ie, I<name=unusual>
is actually interpreted as I<name=*unusual*>).

You can also use the folling ARGUMENT types when searching for field
values in a waypoint:

   field ARGUMENT value

=over

=item field=value

Fuzzy match on value.  The value will be surrounding by wildcards, as
previously discussed so searching for "desc=blowfish" will actually
search the desc field for "*blowfish*".

  waypoint:desc=blowfish

=item field==value

Also as mentioned above, if you don't want to use the automatic wild
card searching discussed above, you can use 2 equal signs to indicate
this.  Thus if you search for "desc=blowfish" it will search the desc
fields for an exact match on the word "blowfish" (IE, must not contain
any other words).

You should use == searches for numeric values too, or else something
like "difficult=3" will also match difficult numbers like 3.5.

  waypoint:desc==blowfish

=item field<value and field>value

Allows you to search on less than or greater than numeric values.  IE:

  waypoint:lon<42.123

or

  waypoint:terrain>2

=item <>

If you use both, it becomes "not equal to".  IE:

  waypoint:type<>unusual

Will list all waypoints B<without> the word I<unusual> in their desc
field.  This is also doing the same automatic wildcard matching like
the = sign does, and thus the above search is actually searching for
"*unusual*".p

=item <=>

This is also "not equal to", but is an exact match and thus:

  waypoint:desc<=>blowfish

will match any type of cache that doesn't have a desc field of
"blowfish".  It will include, in the search, caches with a desc field
of "blowfish dinner" however because it's not an exact match.

=back

END HELP

=head1 AUTHOR

Wes Hardaker (AKA Yamar) <hardaker ATAT users.sourceforge.net> 

=cut

