# Copyright (C) 2007 Wes Hardaker
# License: GNU GPLv2.  See the COPYING file for details.
package GeoDB::Utils;

use strict;

our $VERSION = "0.97";
require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw(GEODEBUG GEOERR GEOWARN GEODUMP
		 init_sql_utils get_last_id select_flat1
		 find_or_make_cache add_or_set_cache_attr
		 find_cache find_caches search_for_caches get_cache_attrs
		 get_caches_from_where_clause format_output
		 parse_coords sprint_coords print_coords
		 calc_distance max min maybe_wrap maybe_to_undef
		 maybe_load copy_file grab_url
		 parse_distance sprint_distance
		 get_macro_parts get_date_stamp);

# database handles
my ($getidh, $getidsh, $newidh, $getattrh, $setattrh, $newattrh, $getattrsh);

my $have_wrap = eval "require Text::Wrap;";
my $have_htmlform = eval "require HTML::FormatText;";
require HTML::TreeBuilder if ($have_htmlform);
import Text::Wrap if ($have_wrap);
use Text::Balanced qw(extract_bracketed);

sub init_sql_utils {
    my $dbh = shift;

    $getidh   = $dbh->prepare_cached('select ident from waypoints
                                       where ident = ?');

    $getidsh  = $dbh->prepare_cached('select ident from waypoints
                                       where ident like ?');

    # attribute handles
    $getattrh = $dbh->prepare_cached('select ident from waypointdata
                                       where ident  = ?
                                         and attrname = ?');
    $newattrh = $dbh->prepare_cached('insert into
                                     waypointdata(ident, attrname, attrvalue)
                                     values(?, ?, ?)');
    $setattrh = $dbh->prepare_cached('update waypointdata
                                         set attrvalue = ?
                                       where ident  = ?
                                         and attrname = ?');

    # all attrs handles
    $getattrsh = $dbh->prepare_cached('select attrname, attrvalue
                                         from waypointdata
                                        where ident  = ?');
}

sub find_cache {
    my ($dbh, $gcid) = @_;

    $getidh->execute($gcid);
    my $row = $getidh->fetchrow_arrayref;
    return $row->[0] if ($row);
}

sub find_caches {
    my ($dbh, $gcid) = @_;

    $getidsh->execute($gcid) || die "$@";
    my $rows = $getidsh->fetchall_arrayref;
    return if (!$rows);
    map { $_ = $_->[0]} @$rows;
    return $rows;
}

sub limit_term {
    return "limit $main::opts{limit}" if ($main::opts{'limit'});
}

sub get_caches_from_where_clause {
    my ($dbh, $clause, $args, $limit) = @_;
    my $limitstring = ($limit ? "limit $limit" : "");

    my $sth = $dbh->prepare_cached("select waypoints.ident from waypoints
                                    left join geocaches
                                           on waypoints.ident = geocaches.ident
                                    left join waymarks
                                           on waypoints.ident = waymarks.ident
                                    left join wigles
                                           on waypoints.ident = wigles.ident
                                    left join geodinings
                                           on waypoints.ident = geodinings.ident
                                     where $clause $limitstring");
    GEODEBUG(2,"searching using: $clause $limitstring\n");
    $sth->execute(@$args) || die "$@";
    my $rows = $sth->fetchall_arrayref;
    return if (!$rows);
    map { $_ = $_->[0]} @$rows;
    return $rows;
}


sub add_or_set_waypoint_attr {
    my ($dbh, $geoid, $attrname, $attrvalue) = @_;

    $getattrh->execute("$geoid", "$attrname");
    my $row = $getattrh->fetchrow_arrayref;
    if ($row) {
	GEODEBUG(6,"updating $geoid $attrname: $attrvalue\n");
	$getattrh->finish();
	$setattrh->execute("$attrvalue", "$geoid", "$attrname") || 
	  GEOERR("failed to update cache #$geoid with $attrvalue -> $attrname");
    } else {
	GEODEBUG(6,"inserting $geoid $attrname: $attrvalue\n");
	$newattrh->execute("$geoid", "$attrname", "$attrvalue") || 
	  GEOERR("failed to insert cache #$geoid with $attrvalue -> $attrname");;
    }
}

sub get_waypoint_attrs {
    my ($dbh, $geoid) = @_;
    $getattrsh->execute("$geoid");
    return $getattrsh->fetchall_arrayref;
}

sub select_flat1 {
    my ($dbh, $statement, @args) = @_;
    my $sth = $dbh->prepare_cached($statement);
    $sth->execute(@args);
    my $rows = $sth->fetchall_arrayref;
    map { $_ = $_->[0]} @$rows;
    return $rows;
}

# wraps and maybe converts from html
sub format_output {
    my ($text, $ishtml) = @_;

    if ($ishtml eq 'True' && $have_htmlform) {
	my $tree = HTML::TreeBuilder->new->parse($text);
	my $form = HTML::FormatText->new(leftmargin => 2, rightmargin => 78);
	return "  " . $form->format($tree);
    }
    return maybe_wrap($text) . "\n";
}

# takes a string with xyzTOKENLEFTblahRIGHTpdq and returns xyz,blah,pdq
# odd numbers in the returned array are the inside of the macros
# a macro followed by a macro will return a '' string in between macro parts
sub get_macro_parts {
    my ($token, $left, $right, $string) = @_;
    my @outparts;
    my $part;
    my $leftregex = "\\$left";
    my $rightregex = "\\$right";
    my $leftcache = "";
    while ($string =~ s/([^$token]*)$token//) {
	my $leftside = $1;
	if ($string =~ /^$leftregex/) {
	    push @outparts, $leftcache . $leftside;
	    $leftcache = "";
	    ($part, $string) = extract_bracketed($string, $left . $right);
	    $part =~ s/^$leftregex//;
	    $part =~ s/$rightregex$//;
	    push @outparts, $part;
	} else {
# 	    if ($#outparts > -1) {
# 		$outparts[$#outparts] .= $token . $leftside;
# 	    } else {
		$leftcache .= $leftside . $token;
# 	    }
	}
    }
    push @outparts, $leftcache . $string;
    return \@outparts;
}

sub maybe_wrap {
    my ($text, $pre1, $pre2) = @_;
    if ($have_wrap) {
        $pre1 ||= "  ";
        $pre2 ||= $pre1;
        return wrap($pre1, $pre2, $text) . "\n";
    }
    return $pre1 . $text;
}

sub get_date_stamp {
    my @tinfo = localtime();
    return sprintf("%04d-%02d-%02dT%02d:%02d:%02d.00000000-07:00",
		   $tinfo[5]+1900,$tinfo[4]+1,$tinfo[3],
		   $tinfo[2], $tinfo[1], $tinfo[0]);
}

#
# distance manipulation routines
#

# takes 90ft, 34km, etc and converts to s. miles
sub parse_distance {
    my $dist = shift;
    return $dist/5280 if ($dist =~ s/ft$//);
    return $dist*0.6213712 if ($dist =~ s/km*s*$//);
    return $dist*0.0006213712 if ($dist =~ s/ms*$//);
    return $dist*1.1507795 if ($dist =~ s/nmis*$//);
    return $dist*0.0005682 if ($dist =~ s/yds*$//);
    return $dist if ($dist =~ s/mis*$//);
    return 0;
}

# takes a distance in miles and converts to a desired type
sub sprint_distance {
    my $dist = shift;
    my $type = shift || $main::opts{'distanceunits'};
    return $dist if (!$type);
    return $dist/0.6213712 if ($type eq 'mi');
    return $dist/0.0006213712 if ($type eq 'm');
    return $dist/1.1507795 if ($type eq 'nmi');
    return $dist/0.0005682 if ($type eq 'yd');
    return $dist; #ugh
}

#
# coordinate versatility
#

sub sprint_coords {
    my ($n, $w) = @_;
    my @ret;
    if ($main::opts{'degrees'}) {
	return ($n, $w);
    }
    foreach my $c ($n, $w) {
	my $hours = int($c);
	my $mins = 60* abs($c-$hours);
	push @ret, sprintf("%s %02.3f",$hours, $mins);
    }
    return @ret;
}

sub print_coords {
    print join(",", sprint_coords(@_));
}

sub parse_coords {
    my ($n, $w) = @_;

    # drop leading spaces
    $n =~ s/^\s+//;
    $w =~ s/^\s+//;

    # and trailing
    chomp($n);
    chomp($w);

    # convert negative letters to a negative sign or drop leading positives
    $n =~ s/^[Ss]/-/;
    $w =~ s/^[Ww]/-/;
    $n =~ s/^[Nn]//;
    $w =~ s/^[Ee]//;

    # parse the north coordinate
    if ($n !~ /^[-\.\d]+$/) {
	if ($n =~ /(-?)\s*(\d+).?\s+(\d+\.\d+)/) {
	    # hh mm
	    $n = $2 + $3/60;
	    $n = - $n if ($1 eq '-');
	} elsif ($n =~ /(-?)\s*(\d+).?\s+(\d+).?\s+(\d+)/) {
	    # hh mm ss
	    $n = $2 + $3/60 + $4/3600;
	    $n = - $n if ($1 eq '-');
	}
    }

    # parse the west coordinate
    if ($w !~ /^[-\.\d]+$/) {
	if ($w =~ /(-?)\s*(\d+).?\s+(\d+\.\d+)/) {
	    $w = $2 + $3/60;
	    $w = - $w if ($1 eq '-');
	} elsif ($w =~ /(-?)\s*(\d+).?\s+(\d+).?\s+(\d+)/) {
	    # hh mm ss
	    $w = $2 + $3/60 + $4/3600;
	    $w = - $w if ($1 eq '-');
	}
    }

    return ($n, $w);
}

#
# Coordinate manipulation routines
#

my $r = 3963.1; # miles
my $pi = atan2(1,1) * 4;

# cos(D/$r)
# args: n1, w1, n2, w2
sub calc_distance {
    my @a = @_;
    map { $_ = $_*($pi/180); } @a;   # map to radians
    return acos(cos($a[0])*cos($a[1])*cos($a[2])*cos($a[3]) +
		cos($a[0])*sin($a[1])*cos($a[2])*sin($a[3]) +
		sin($a[0])*sin($a[2])) * $r;
}

sub course_degs {
    my @a = @_;
    map { $_ = $_*($pi/180); } @a;   # map to radians

      # acos(sin($a[0])*sin($a[2])+cos($a[0])*cos($a[2])*cos($a[1]-$a[3]));
    my $d=
      2*asin(sqrt((sin(($a[0]-$a[2])/2))**2 + 
		  cos($a[0])*cos($a[2])*(sin(($a[1]-$a[3])/2))**2));
    my $course;

    map { print "  $_\t",sin($_),"\t",cos($_),"\n"; } @a;

    print "d: $d\n";

    if (sin($a[3]-$a[1]) < 0) {
	$course = acos((sin($a[2])-sin($a[0])*cos($d))/(sin($d)*cos($a[0])));
    } else {
	$course =
	  2*$pi-acos((sin($a[2])-sin($a[0])*cos($d))/(sin($d)*cos($a[0])))
    }
    $course = $course * 180 / $pi;
    print "course: $course\n";
    return $course;
}

#
# MATH Routines
#

sub acos {
   my($x) = @_;
   return 0 if ($x eq 1);
   my $ret = atan2(sqrt(1 - $x**2), $x);
   return $ret;
}

sub asin {
   my($x) = @_;
   my $ret = atan2($x, sqrt(1 - $x**2));
   return $ret;
}

sub max {
    return ($_[0] > $_[1]) ? $_[0] : $_[1];
}

sub min {
    return ($_[0] < $_[1]) ? $_[0] : $_[1];
}

#
# simple file copying
#
sub copy_file {
    my ($in, $out) = @_;
    open(ICOPY, $in) || GEOERR("can't copy to file $out");
    open(OCOPY, ">$out") || GEOERR("can't copy to file $out");
    my $buf;
    while (read(ICOPY,$buf,4096) > 0) {
	print OCOPY $buf;
    }
    close(ICOPY);
    close(OCOPY);
}

# pull data from a URL
sub grab_url {
    my ($url, $to, $overwrite) = @_;
    return if (-f $to && !$overwrite);
    require LWP::UserAgent;
    my $ua = LWP::UserAgent->new;
    $url =~ s/^\s*//s;
    $url =~ s/\s*$//s;
    my $res = $ua->get($url, ':content_file' => $to);
    if (!$res->is_success) {
	GEOERR("Grabbing a url failed: " . $res->status_line);
    }
}

sub maybe_to_undef {
    return undef if (!defined($_[0]) || $_[0] eq '');
    return $_[0];
}

my %tabcache;
my $last_insert_row_id;
sub get_last_id {
    my ($dbh, $sth, $table, $col) = @_;
    if (exists($sth->{'mysql_insertid'})) {
	# MYSQL specific method
	$last_insert_row_id = $sth->{'mysql_insertid'};
	return $last_insert_row_id;
    } elsif (exists($sth->{pg_oid_status})) {
	# Postgres specific method
	$col = 'id' if (!$col);
	my $resh = $tabcache{$table} ||
	  $dbh->prepare("select $col from $table where oid = ?");
	$resh->execute($sth->{pg_oid_status});
	my $results = $resh->fetchrow_arrayref;
	$last_insert_row_id = $results->[0];
	$resh->finish;
	return $last_insert_row_id;
    }
    # XXX: handle others
}

#
# debugging and global error handling
#

sub GEODEBUG {
    my $level = shift;
    print "$level:" . (" " x ($level > 9 ? 9 : $level)),@_
      if ($main::opts{'debug'} && $main::opts{'debug'} >= $level);
}

sub GEOWARN {
    print STDERR "\nWARNING:\n  ",@_,"\n";
    if ($main::opts{'debug'}) {
	print STDERR "Stack Trace:\n";
	for (my $i = 1; $i <= 20; $i++) {
	    my @everything = caller($i);
	    printf STDERR ("  %-60s line: %d\n",
			   $everything[0] . "::" .  $everything[3],
			   $everything[2]) if ($everything[0]);
	}
    }
}

sub GEOERR {
    print STDERR "\n\nFatal Error:\n  ",@_,"\n\n";
    if ($main::opts{'debug'}) {
	print STDERR "Stack Trace:\n";
	for (my $i = 1; $i <= 20; $i++) {
	    my @everything = caller($i);
	    printf STDERR ("  %-60s line: %d\n",
			   $everything[0] . "::" .  $everything[3],
			   $everything[2]) if ($everything[0]);
	}
    }
    exit(1);
}

sub GEODUMP {
    require Data::Dumper;
    import Data::Dumper;
    print Dumper(@_);
}

# conditional perl module loading
sub maybe_load {
    my ($module, $errtxt, $importspecs) = @_;
    my $haveit = eval "require $module;";
    if (!$haveit) {
	return 0 if ($errtxt eq 'noerror');
	$errtxt ||= "Failed to load perl module $module\n  You need this to perform a function you just requested."; 
	GEOERR($errtxt);
    }
    eval "import $module $importspecs;";
    return 1;
}

1;
