# Copyright (C) 2007 Wes Hardaker
# License: GNU GPLv2.  See the COPYING file for details.
package GeoDB::Waypoint;

use strict;
use GeoDB::Utils;
require GeoDB::Waypoint::Geocache;

our $VERSION = "0.97";
require Exporter;
our @EXPORT = qw(register_waypoint_type
		 get_registered_cols_and_map get_waypoint_type_cols_and_map
		 create_waypoint_handles);
our @ISA = qw(Exporter);

# database handles
my ($getwaypointh, $getidh, $newwidh, $newgidh, $updatewh,
    $getattrh, $setattrh, $newattrh, $getattrsh, $attrdelh,
    $gettagsh, $settagh, $newtagh,
    $getratingsh, $setratingh, $newratingh,
    $logdelh, $logaddh, $logselh, $logselfh, 

    # registered objects
    $getdatah, $classcols, $classmap);

# 'ident' not listed intentionally
# descr intentionally different than the desc gpx field (desc is reserved sql)
our @waypoint_dbcols = qw(
			  descr
			  lat
			  lon
			  sym

			  time
			  type
			  subtype
			  url
			  urlname
			  cmt
			 );

# 'ident' not listed intentionally
our @waypoint_classcols = qw(
			     desc
			     lat
			     lon
			     sym

			     time
			     type
			     subtype
			     url
			     urlname
			     cmt
			     );

our %registered_types;
my $delfromdbh;

sub init {
    my $dbh = shift;

    return if (!$dbh);

    # geocache handles
    $getidh   = $dbh->prepare_cached('select ident from waypoints
                                       where ident = ?');

    $newwidh   = $dbh->prepare_cached('insert into waypoints(' .
				      join(", ", @waypoint_dbcols) .
				      ', ident) values(' .
				      ("?, " x ($#waypoint_dbcols + 1))." ?)");

    $updatewh   = $dbh->prepare_cached('update waypoints set ' .
				       join(" = ?, ", @waypoint_dbcols) .
				       ' = ? where ident = ?');

    $getwaypointh =
      $dbh->prepare_cached('select ' .
			   join(", ", @waypoint_dbcols) .
			   ' from waypoints
                            where waypoints.ident = ?');

    # log handles
    $logdelh = $dbh->prepare_cached('delete from logs where id = ?');
    $logaddh = $dbh->prepare_cached('insert into logs (id, ident, date, type,
                                                       finder_id, finder,
                                                       text_encoded, text)
                                                values(?,?,?,?,?,?,?,?)');
    $logselh = $dbh->prepare_cached('select id, date, type,
                                            finder_id, finder,
                                            text_encoded, text
                                       from logs
                                      where ident = ?');

    $logselfh = $dbh->prepare_cached('select id, date, type,
                                             finder_id, finder,
                                             text_encoded, text
                                        from logs
                                       where ident = ? and finder = ?');

    # attribute handles
    $attrdelh = $dbh->prepare_cached('delete from waypointdata
                                      where ident = ? and attrname = ?');
    $getattrh = $dbh->prepare_cached('select ident from waypointdata
                                       where ident  = ?
                                         and attrname = ?');
    $newattrh = $dbh->prepare_cached('insert into
                                     waypointdata(ident, attrname, attrvalue)
                                     values(?, ?, ?)');
    $setattrh = $dbh->prepare_cached('update waypointdata
                                         set attrvalue = ?
                                       where ident  = ?
                                         and attrname = ?');

    # all attrs handles
    $getattrsh = $dbh->prepare_cached('select attrname, attrvalue
                                         from waypointdata
                                        where ident  = ?');

    # all tags handles
    $gettagsh = $dbh->prepare_cached('select tagname, ismine, needspublished,
                                             count, datefetched, datepublished
                                        from tags
                                       where ident  = ?');

    $settagh = $dbh->prepare_cached('update tags
                                         set needspublished = 1,
                                             ismine = ?
                                       where ident  = ? and tagname = ?');

    $newtagh = $dbh->prepare_cached('insert into tags
                                       (ident, tagname, ismine, needspublished)
                                       values(?,?,1,1)');

    # all ratings handles
    $getratingsh = $dbh->prepare_cached('select ratingname, ratingvalue,
                                             othersrating, needspublished,
                                             count, datefetched, datepublished
                                        from ratings
                                       where ident  = ?');

    $setratingh = $dbh->prepare_cached('update ratings
                                         set needspublished = 1,
                                             ratingvalue = ?
                                       where ident  = ? and ratingname = ?');

    $newratingh = $dbh->prepare_cached('insert into ratings
                                       (ident, ratingname, ratingvalue,
                                        needspublished)
                                       values(?,?,?,1)');

    $delfromdbh = $dbh->prepare_cached('delete from ? where ident = ?');

    get_data_statement(undef, $dbh);

    for (my $i = 0; $i <= $#waypoint_classcols; $i++) {
	$classmap->{$waypoint_classcols[$i]} = $waypoint_dbcols[$i];
    }
    $classmap->{'ident'} = 'waypoints.ident';

    return 1;
}

sub create_waypoint_handles {
    my ($dbh, $table, $dbcols) = @_;

    return if (!$dbh);

    my $getidh = $dbh->prepare_cached("select ident from $table
                                       where ident = ?");

    my $newwayidh   = $dbh->prepare_cached("insert into $table(" .
					   join(", ", @$dbcols) .
					   ', ident) values(' .
					   ("?, " x ($#$dbcols + 1)). " ?)");

    my $updatewh   = $dbh->prepare_cached("update $table set " .
					  join(" = ?, ", @$dbcols) .
					  ' = ? where ident = ?');

    return ($getidh, $newwayidh, $updatewh);
}


sub new {
    my $type = shift;
    my ($class) = ref($type) || $type;
    my $self;
    if (ref($_[0]) eq 'HASH') {
	$self = $_[0];
    } else {
	my %self = @_;
	$self = \%self;
    }
    return if (!$self->{'data'}{'ident'}); # required field
    bless($self, $class);
}

sub clone {
    my $self = shift;
    my $clone;
    %{$clone} = %{$self};
    $clone->{'data'} = {};
    %{$clone->{'data'}} = %{$self->{'data'}};
    if (exists($self->{'attributes'})) {
	$clone->{'attributes'} = {};
	%{$clone->{'attributes'}} = %{$self->{'attributes'}}
    }
    bless($clone, ref($self));
}

sub clone_clean {
    my $self = shift;
    my $clone;
    $clone->{'data'} = {ident => $self->{'data'}{'ident'}};
    bless($clone, ref($self));
}

sub ident {
    return $_[0]->{'data'}{'newident'} || $_[0]->{'data'}{'ident'};
}

sub get_data_statement {
    my ($self, $dbh) = @_;

    # return previously constructed handle
    return ($getdatah, $classcols) if ($getdatah);

    my $tables = "waypoints";
    my @classcols = @waypoint_classcols;
    my @dbcolumns = @waypoint_dbcols;

    # create a new handle merging all known types together (ugh).

    foreach my $type (keys(%registered_types)) {
	$tables .= "
    left join $registered_types{$type}{'table'}
          on   waypoints.ident = $registered_types{$type}{'table'}.ident
";
	push @dbcolumns, @{$registered_types{$type}{'dbcolumns'}};
	push @classcols, @{$registered_types{$type}{'classcolumns'}};
    }

    #
    my $getdatahstr =  "select " . join(", ", @dbcolumns) . " from $tables 
                         where waypoints.ident = ?";
    $getdatah = $dbh->prepare_cached($getdatahstr);

    $classcols = \@classcols;

    return ($getdatah, $classcols);
}

sub register_waypoint_type {
    my ($type, $classname, $table, $dbcolumns, $classcolumns) = @_;

    my %typeclassmap;

    for (my $i = 0; $i <= $#$classcolumns; $i++) {
	$classmap->{$classcolumns->[$i]} = $dbcolumns->[$i];
	$typeclassmap{$classcolumns->[$i]} = $dbcolumns->[$i];
    }

    # merge in main DB table
    for (my $i = 0; $i <= $#waypoint_classcols; $i++) {
	$typeclassmap{$waypoint_classcols[$i]} = $waypoint_dbcols[$i];
    }

    $registered_types{$type} = 
      {
       classname    => $classname,
       table   	    => $table,
       dbcolumns    => $dbcolumns,
       classcolumns => $classcolumns,
       classmap     => \%typeclassmap,
      };

}

sub get_waypoint_type_cols_and_map {
    my ($type) = @_;
    return (['ident', @waypoint_dbcols,
	     @{$registered_types{$type}{'dbcolumns'}}],
	    $registered_types{$type}{'classmap'});
}

sub get_registered_cols_and_map {
    return (['ident', @$classcols], $classmap);
}

# load the basic data
sub load_waypoint {
    my ($self) = @_;

    GEODEBUG(7,"loading data for $self->{'data'}{'ident'}\n");
    my ($sth, $classcols) = $self->get_data_statement();
    $sth->execute($self->{'data'}{'ident'});
    my $row = $sth->fetchrow_arrayref();
    for (my $i = 0; $i <= $#$classcols; $i++) {
	$self->{'data'}{$classcols->[$i]} = $row->[$i];
    }
    $sth->finish();
}

# load the cache from the DB
#  this does postive loading: options are used to specify what *to* load
sub load_parts {
    my ($self, $options) = @_;

    # pull from db
    GEODEBUG(6,"loading cache $self->{'data'}{'ident'}\n");
    $self->load_logs($options)          if ($options->{'logs'});
    $self->load_attributes($options)    if ($options->{'attributes'});
    $self->load_tags($options)          if ($options->{'tags'});
    $self->load_ratings($options)       if ($options->{'ratings'});

    return if (exists($self->{'loadeddata'}));
    $self->{'loadeddata'} = 1;

    $self->load_waypoint();

    $self->morph_to_specific();
    $self->load_sub($options);
}

# load the cache from the DB
#  this does negative loading: options are used to specify what *not* to load
#  (use load_parts instead to avoid loading what you don't want when future new
#   types of data pieces evolve)
sub load {
    my ($self, $options) = @_;

    # pull from db
    GEODEBUG(6,"loading cache $self->{'data'}{'ident'}\n");
    $self->load_logs($options);
    $self->load_attributes($options);
    $self->load_tags($options);
    $self->load_ratings($options);

    return if (exists($self->{'loadeddata'}));
    $self->{'loadeddata'} = 1;

    $self->load_waypoint();

    $self->morph_to_specific();
    $self->load_sub($options);
}

sub morph_to_specific {
    my ($self) = @_;
    # XXX: use the registered type instead
    if (exists($registered_types{$self->{'data'}{'type'}})) {
	bless($self, $registered_types{$self->{'data'}{'type'}}{'classname'});
    }
}

sub load_sub {
}

########################################
# TAGS
#

sub load_tags {
    # hook for subs if the need it.
    my ($self, $options) = @_;

    my $handle;

    return if ($options->{'notags'} || $self->{'loadedtags'});
    $self->{'loadedtags'} = 1;

    # load the logs XXX: make optional loading
    GEODEBUG(7,"loading tags for $self->{'data'}{'ident'}\n");

    my %alltags;
    $gettagsh->execute($self->{'data'}{'ident'});
    my $tags = $gettagsh->fetchall_arrayref();
    map {
	$alltags{$_->[0]} = { ismine => $_->[1],
			      needspublished => $_->[2],
			      count => $_->[3],
			      datafetched => $_->[4],
			      datepublished => $_->[5]
			    };
    } @$tags;
    $self->{'tags'} = \%alltags;
    $gettagsh->finish();
}

sub set_tag {
    my ($self, $tagname) = @_;
    if (exists($self->{'tags'}{$tagname})) {
	if (! $self->{'tags'}{$tagname}{'ismine'}) {
	    $settagh->execute(1, $self->{'data'}{'ident'}, $tagname);
	    $self->{'tags'}{$tagname}{'ismine'} = 1;
	    $self->{'tags'}{$tagname}{'needspublished'} = 1;
	}
    } else {
	$newtagh->execute($self->{'data'}{'ident'}, $tagname);
	$self->{'tags'}{$tagname} =
	  {ismine => 1, needspublished => 1, count => 0}
    }
}

sub remove_tag {
    my ($self, $tagname) = @_;
    if (exists($self->{'tags'}{$tagname})) {
	$settagh->execute(0, $self->{'data'}{'ident'}, $tagname);
	$self->{'tags'}{$tagname}{'needspublished'} = 1;
	$self->{'tags'}{$tagname}{'ismine'} = 0;
    }
}

sub sshow_tags {
    my ($self) = @_;
    my @tags = keys(%{$self->{'tags'}});
    my (@mytags, @theirtags);
    my $out;
    $self->load_parts({tags => 1});
    foreach my $tag (sort keys(%{$self->{'tags'}})) {
	if ($self->{'tags'}{$tag}{'ismine'}) {
	    push @mytags, $tag;
	}
	if ($self->{'tags'}{$tag}{'count'} > 0) {
	    push @theirtags, "$tag($self->{'tags'}{$tag}{'count'})";
	}
    }
    $out .= "Your Tags: " . join(", ", @mytags)  . "\n" if ($#mytags > -1);
    $out .= "Other Tags: " . join(", ", @theirtags)  . "\n" if ($#theirtags > -1);
    return $out;
}

sub show_tags {
    my $self = shift;
    print $self->sshow_tags(@_);
}


########################################
# RATINGS
#

sub load_ratings {
    # hook for subs if the need it.
    my ($self, $options) = @_;

    my $handle;

    return if ($options->{'noratings'} || $self->{'loadedratings'});
    $self->{'loadedratings'} = 1;

    # load the logs XXX: make optional loading
    GEODEBUG(7,"loading ratings for $self->{'data'}{'ident'}\n");

    my %allratings;
    $getratingsh->execute($self->{'data'}{'ident'});
    my $ratings = $getratingsh->fetchall_arrayref();
    map {
	$allratings{$_->[0]} = {
				value => $_->[1],
				othersrating => $_->[2],
				needspublished => $_->[3],
				count => $_->[4],
				datafetched => $_->[5],
				datepublished => $_->[6]
			    };
    } @$ratings;
    $self->{'ratings'} = \%allratings;
    $getratingsh->finish();
}

sub set_rating {
    my ($self, $ratingname, $value) = @_;
    if (exists($self->{'ratings'}{$ratingname})) {
	if (exists($self->{'ratings'}{$ratingname}{'value'})) {
	    $setratingh->execute($value, $self->{'data'}{'ident'}, $ratingname);
	    $self->{'ratings'}{$ratingname}{'value'} = $value;
	    $self->{'ratings'}{$ratingname}{'needspublished'} = 1;
	}
    } else {
	$newratingh->execute($self->{'data'}{'ident'}, $ratingname, $value);
	$self->{'ratings'}{$ratingname} =
	  {value => $value, needspublished => 1, count => 0}
    }
}

sub remove_rating {
    my ($self, $ratingname) = @_;
    if (exists($self->{'ratings'}{$ratingname})) {
	$setratingh->execute(undef, $self->{'data'}{'ident'}, $ratingname);
	$self->{'ratings'}{$ratingname}{'needspublished'} = 1;
	delete $self->{'ratings'}{$ratingname}{'value'};
    }
}

sub sshow_ratings {
    my ($self) = @_;
    my @ratings = keys(%{$self->{'ratings'}});
    my (@myratings, @theirratings);
    my $out;
    $self->load_parts({ratings => 1});
    foreach my $rating (sort keys(%{$self->{'ratings'}})) {
	if (exists($self->{'ratings'}{$rating}{'value'})) {
	    push @myratings, "$rating=$self->{'ratings'}{$rating}{'value'}";
	}
	if ($self->{'ratings'}{$rating}{'count'} > 0) {
	    push @theirratings, "$rating($self->{'ratings'}{$rating}{'count'})=$self->{'ratings'}{$rating}{'othersrating'}";
	}
    }
    $out .= "Your Ratings: " . join(", ", @myratings)  . "\n" if ($#myratings > -1);
    $out .= "Other Ratings: " . join(", ", @theirratings)  . "\n" if ($#theirratings > -1);
    return $out;
}

sub show_ratings {
    my $self = shift;
    print $self->show_ratings(@_);
}


########################################
# ATTRIBUTES
#
sub load_attributes {
    # hook for subs if the need it.
    my ($self, $options) = @_;

    my $handle;

    return if ($options->{'noattributes'} || $self->{'loadedattributes'});
    $self->{'loadedattributes'} = 1;

    # load the logs XXX: make optional loading
    GEODEBUG(7,"loading attributes for $self->{'data'}{'ident'}\n");

    my %attributes;
    my $attr;
    $getattrsh->execute($self->{'data'}{'ident'});
    my $attrs = $getattrsh->fetchall_arrayref();
    map { $attributes{$_->[0]} = $_->[1]} @$attrs;
    $self->{'attributes'} = \%attributes;
    $getattrsh->finish();
}

sub sshow_attributes {
    my ($self) = @_;
    my @attrkeys = keys(%{$self->{'attributes'}});
    my $out;
    if ($#attrkeys > -1) {
	$out = sprintf("Attributes: \n");
	foreach my $attr (sort @attrkeys) {
	    $out .= 
	      sprintf("  %-20s %s\n", $attr, $self->{'attributes'}{$attr});
	}
    }
    return $out;
}

sub show_attributes {
    my ($self) = @_;
    print $self->sshow_attributes();
}

sub set_attribute {
    my ($self, $attrname, $value) = @_;
    $self->{'attributes'}{$attrname} = $value;
}

sub get_attribute {
    my ($self, $attrname, $nosubstitute, $loopcache) = @_;

    return if (!exists($self->{'attributes'}{$attrname}));
    # get the attribute value

    # avoid recursive loops
    if (exists($loopcache->{$attrname})) {
	GEOWARN("Warning: recursive loop in $self->{data}{ident}: $attrname");
	return;
    }
    $loopcache->{$attrname} = 1;
    GEODEBUG(5,"getting attribute $attrname from $self->{data}{ident}\n");

    my $value = $self->{'attributes'}{$attrname};

    # return the calculated replacement by default
    # A=math:1+2        should => 3
    # B=replace:D A C   should => D 3 C
    unless ($nosubstitute) {
	if ($value =~ s/^(math|substitute|replace)://) {
	    my $operation = $1;
	    $value =~ s/PLUS/+/g;
	    $value =~ s/MINUS/-/g;
	    $value =~ s/\(([a-zA-Z]\w*)\)/my $newv = $self->get_attribute($1,$nosubstitute,$loopcache); (defined($newv)) ? $newv : "($1)"/ge;
	    $value = eval($value) if ($operation eq 'math');
	}
    }
    delete $loopcache->{$attrname};
    GEODEBUG(5,"done getting attribute $attrname from $self->{data}{ident}\n");
    return $value;
}

sub save_attributes {
    my ($self) = @_;
    my $ident = $self->{'data'}{'ident'};
    # save attributes
    foreach my $attrid (keys(%{$self->{'attributes'}})) {
	GEODEBUG(6,"saving attr $attrid\n");
	my $attr = $self->{'attributess'}{$attrid};
	$attrdelh->execute($ident, $attrid);
	$newattrh->execute($ident, $attrid, $self->{'attributes'}{$attrid});
	GEODEBUG(6,"saved $attrid => $self->{'attributes'}{$attrid}\n");
	if ($newattrh->errstr) {
	    print "Error saving attr $attrid: " . $newattrh->errstr . "\n";
	}
    }
}

########################################
# LOGS
#
sub load_logs {
    my ($self, $options) = @_;

    my $handle;

    return if ($options->{'nologs'} || $self->{'loadedlogs'});
    $self->{'loadedlogs'} = 1 if (!$options->{'logfinder'});

    # load the logs XXX: make optional loading
    GEODEBUG(7,"loading logs for $self->{'data'}{'ident'}\n");
    if ($options->{'logfinder'}) {
	$handle = $logselfh;
	$handle->execute($self->{'data'}{'ident'}, $options->{'logfinder'});
    } else {
	$handle = $logselh;
	$handle->execute($self->{'data'}{'ident'});
    }
    my $log;
    my $logs = {};
    while ($log = $handle->fetchrow_arrayref()) {
	$logs->{$log->[0]} = 
	  {
	   id => $log->[0],
	   'date' => $log->[1],
	   'type' => $log->[2],
	   'finder_id' => $log->[3],
	   'finder' => $log->[4],
	   'text_encoded' => $log->[5],
	   'text' => $log->[6],
	  };
    }
    $self->{'logs'} = $logs;
    $logselh->finish();
}

sub sshow_logs {
    my ($self) = @_;
    my $out;
    foreach my $logid (sort keys(%{$self->{'logs'}})) {
	my $log = $self->{'logs'}{$logid};
	$out .= 
	  sprintf("\nLog #$log->{id} by $log->{'finder'}: ($log->{type})\n");
	$out .= format_output($log->{'text'});
    }
    return $out;
}

sub show_logs {
    my $self = shift;
    print $self->sshow_logs(@_);
}

sub typedata_exists {
    return 0;
}

sub waypointdata_exists {
    my ($self) = @_;
    $getidh->execute($self->{'data'}{'ident'});
    my $row = $getidh->fetchrow_arrayref();
    $getidh->finish();
    return 1 if $row;
    return 0;
}

# if it already exists in the db, return 1
sub itexists {
    my $self = shift;
    return $self->waypointdata_exists(@_) || $self->typedata_exists(@_);
}

# save the current cache to the DB (either update or insert a new row)
sub save {
    my $self = shift;
    my $modwh = $newwidh;
    # save to db

    if ($self->itexists()) {
	GEODEBUG(5,"updating waypoint $self->{data}{ident}\n");
	$modwh = $updatewh;
    } else {
	GEODEBUG(5,"inserting waypoint $self->{data}{ident}\n");
    }
    my $data = $self->{'data'};
    my $ident = $data->{'ident'};
    $modwh->execute($data->{'desc'},
		    $data->{'lat'},
		    $data->{'lon'},
		    $data->{'sym'},

		    $data->{'time'},
		    $data->{'type'},
		    $data->{'subtype'},
		    $data->{'url'},
		    $data->{'urlname'},
		    $data->{'cmt'},

		    # must be last
		    $ident);

    if ($modwh->errstr) {
	GEOERR("Error saving $ident: " . $modwh->errstr . "\n");
    }

    # save logs
    foreach my $logid (keys(%{$self->{'logs'}})) {
	my $log = $self->{'logs'}{$logid};
	$logdelh->execute($logid);
	$logaddh->execute($logid, $ident,
			  $log->{'date'}, $log->{'type'},
			  $log->{'finder_id'}, $log->{'finder'},
			  $log->{'text_encoded'}, $log->{'text'});
	GEODEBUG(6,"saving log $logid\n");
	if ($logaddh->errstr) {
	    print "Error saving log $logid: " . $logaddh->errstr . "\n";
	}
    }
    $self->save_attributes($ident);
}

######################################################################
# Summary and Printing

sub sprint_summary_line {
    my ($self, $count, $format) = @_;

    my @args;
    $self->{'data'}{'count'} = $count;
    return $self->replace_text($format);

#     while ($format =~ s/\%{(?:([\.\d]+):|)([^\%]+)}/"%" . ($1?"-$1":"") . "s"/e) {
# 	push @args, ($2 eq 'count' ? $$count :
# 		     ($self->{'data'}{$2} || $self->{'attributes'}{$2}));
#     }
#     sprintf($format, @args);
}

sub print_summary_line {
    my $self = shift;
    print $self->sprint_summary_line(@_);
}

sub ssummary {
    my ($self, $count, $options) = @_;

    my $summary = ($options->{'format'} ? "$options->{'format'}\n"
		   : (exists($self->{'data'}{'distance'}) ?
		      "%{3:count} %{4.4:distance} %{8:ident} %{66:urlname}\n" :
		      "%{3:count} %{8:ident} %{66:urlname}\n"));
    return
      $self->sprint_summary_line($count, ($options->{'format'} ? "$options->{'format'}\n" : "%{3:count} %{8:ident} %{66:urlname}\n"));
}

sub summary {
    my ($self) = shift;
    print $self->ssummary(@_);
}	

# display the geocache in all it's glory.
sub sdisplay {
    my ($self, $dologs) = @_;
    my $data = $self->{'data'};
    my $out;
    $out .= 
      sprintf("%-8s   %-31s Location:  %s, %s\n",
	      $self->ident(), $data->{'type'} . " / " . $data->{'subtype'},
	      sprint_coords($data->{'lat'}, $data->{'lon'}));
    $out .= sprintf("Name:      %s\n", $self->{'data'}{'urlname'});
    my $time = $data->{'time'};
    $time =~ s/T.*//;
    $out .= sprintf("Created:   %s\n", $time,);
    $out .= sprintf("URL:       %s\n", $data->{'url'});
    $out .= sprintf("Descript:  %s\n", $data->{'desc'});
    $out .= $self->sshow_attributes();
    $out .= $self->sshow_tags();
    $out .= $self->sshow_ratings();
    $out .= $self->sshow_logs() if ($dologs);
    return $out;
}

# display the geocache in all it's glory.
sub display {
    my ($self, $dologs) = @_;
    print $self->sdisplay($dologs);
}

sub edit {
    # returns a qwizard primary usable for editing
    my ($self, $qw) = @_;
    my $primary = 
      { title => 'Editing Geocache @ident@',
	questions => [ { type => 'label',
			 text => 'Geocache ID:',
			 values => '@ident@' }
		     ],
	actions => [sub {
			my $self = $_[$#_];
			foreach my $i (@waypoint_classcols) {
			    $self->{$i} = qwparam($i);
			}
			$self->save();
		    }, $self],
      };
    foreach my $i (@waypoint_classcols) {
	push @{$primary->{'questions'}},
	  { type => 'text',
	    name => $i,
	    default => $self->{$i}
	  }
    }
    return $primary;
}

sub replace_text {
    my ($self, $value) = @_;

    # special keys
    $value =~ s/\%{attr:([^\}]+)}/$self->{'attributes'}{$1}/eg;
    if ($value =~ /tags:/) {
	print "hmmm: $value\n";
    }
    $value =~ s/\%{tags:}/$self->sshow_tags()/eg;
    $value =~ s/\%{ratings:}/$self->sshow_ratings()/eg;
    $value =~ s/\%{attributes:}/$self->sshow_attributes()/eg;

    # general other text, possibly number.number formatted.
    my @args;
    while ($value =~ s/\%{(?:([\.\d]+):|)([^\%]+)}/"%" . ($1?"-$1":"") . "s"/e) {
	push @args, $self->{'data'}{$2} || $self->{'attributes'}{$2};
    }
    $value = sprintf($value, @args);

    # special replacements
    $value =~ s/\\n/\n/g;

    return $value;
}

sub remove_from_db {
    my $self = shift;
    $self->load({'nologs' => 1, 'notags' => 1, 'noattributes' => 1, 'noratings' => 1});
    $self->remove_from_specific_tables(@_);
    $self->remove_from_common_tables(@_);
}

sub remove_from_a_table_by_ident {
    my ($self, $tablename) = @_;
    $delfromdbh->execute($tablename, $self->{'data'}{'ident'});
}

sub remove_from_common_tables {
    my $self = shift;
    $self->remove_from_a_table_by_ident('logs');
    $self->remove_from_a_table_by_ident('ratings');
    $self->remove_from_a_table_by_ident('tags');
    $self->remove_from_a_table_by_ident('sets');
    $self->remove_from_a_table_by_ident('waypointdata');
    $self->remove_from_a_table_by_ident('waypoints');
}

sub parse_coordinates {
    my $self = shift;
    ($self->{'data'}{'lat'}, $self->{'data'}{'lon'}) =
      parse_coords($self->{'data'}{'lat'}, $self->{'data'}{'lon'});
}

# override by child
sub remove_from_specific_tables {
}
1;
