# Copyright (C) 2007 Wes Hardaker
# License: GNU GPLv2.  See the COPYING file for details.
package GeoDB::Waypoint::Geocache;

use strict;
use GeoDB::Utils;
use GeoDB::Waypoint;

our $VERSION = "0.97";
our @ISA=qw(GeoDB::Waypoint);

# database handles
my ($getcacheh, $getidh, $newwidh, $newgidh, $updategh, $updatewh,
    $getattrh, $setattrh, $newattrh, $getattrsh,
    $logdelh, $logaddh, $logselh, $logselfh);

# 'ident' not listed intentionally
our @geocache_dbcols = qw(
			  groundspeak_id

			  groundspeak_available
			  groundspeak_archived
			  groundspeak_name
			  groundspeak_placed_by
			  groundspeak_owner_id

			  groundspeak_owner_name
			  groundspeak_type
			  groundspeak_container
			  groundspeak_difficulty
			  groundspeak_terrain

			  groundspeak_country
			  groundspeak_state
			  groundspeak_short_description_html
			  groundspeak_short_description_text
			  groundspeak_long_description_html

			  groundspeak_long_description_text
			  groundspeak_encoded_hints
			 );

# 'ident' not listed intentionally
our @geocache_class_cols = qw(
			      id

			      available
			      archived
			      groundspeak_name
			      groundspeak_placed_by
			      groundspeak_owner_id

			      groundspeak_owner_name
			      groundspeak_type
			      groundspeak_container
			      groundspeak_difficulty
			      groundspeak_terrain

			      groundspeak_country
			      groundspeak_state
			      groundspeak_short_description_html
			      groundspeak_short_description
			      groundspeak_long_description_html

			      groundspeak_long_description
			      groundspeak_encoded_hints
			     );

our @geocache_waypoint_cols = (
			       @GeoDB::Waypoint::waypoints_dbcols,
			       @geocache_dbcols
			      );

sub init {
    my $dbh = shift;

    return if (!$dbh);

    # geocache handles
    $getidh   = $dbh->prepare_cached('select ident from geocaches
                                       where ident = ?');

    $newgidh   = $dbh->prepare_cached('insert into geocaches(' .
				      join(", ", @geocache_dbcols) .
				      ', ident) values(' .
				      ("?, " x ($#geocache_dbcols + 1)). " ?)");

    $updategh   = $dbh->prepare_cached('update geocaches set ' .
				       join(" = ?, ", @geocache_dbcols) .
				       ' = ? where ident = ?');

    $getcacheh =
      $dbh->prepare_cached('select ' .
			   join(", ", @GeoDB::Waypoint::waypoints_dbcols,
				@geocache_dbcols) .
			   ' from waypoints
                        left join geocaches
                               on waypoints.ident = geocaches.ident
                            where waypoints.ident = ?');

    GeoDB::Waypoint::register_waypoint_type('Geocache', 'GeoDB::Waypoint::Geocache',
			   'geocaches',
			   \@geocache_dbcols, \@geocache_class_cols);

    return 1;
}

# if it already exists in the db, return 1
sub typedata_exits {
    my ($self) = @_;
    $getidh->execute($self->{'data'}{'ident'});
    my $row = $getidh->fetchrow_arrayref();
    $getidh->finish();
    return 1 if $row;
    return 0;
}

# save the current cache to the DB (either update or insert a new row)
sub save {
    my $self = shift;
    my $modgh = $newgidh;
    # save to db

    $self->SUPER::save(@_);

    if ($self->typedata_exits()) {
	GEODEBUG(6,"updating geocache $self->{data}{ident}\n");
	$modgh = $updategh;
    } else {
	GEODEBUG(6,"inserting geocache $self->{data}{ident}\n");
    }
    my $data = $self->{'data'};

    $modgh->execute(maybe_to_undef($data->{'id'}),
		    $data->{'available'},
		    $data->{'archived'},
		    $data->{'groundspeak_name'},
		    $data->{'groundspeak_placed_by'},
		    maybe_to_undef($data->{'groundspeak_owner_id'}),
		    $data->{'groundspeak_owner_name'},
		    $data->{'groundspeak_type'},
		    $data->{'groundspeak_container'},
		    maybe_to_undef($data->{'groundspeak_difficulty'}),
		    maybe_to_undef($data->{'groundspeak_terrain'}),
		    $data->{'groundspeak_country'},
		    $data->{'groundspeak_state'},
		    $data->{'groundspeak_short_description_html'},
		    $data->{'groundspeak_short_description'},
		    $data->{'groundspeak_long_description_html'},
		    $data->{'groundspeak_long_description'},
		    $data->{'groundspeak_encoded_hints'},
		    $data->{'ident'},
		  );
    if ($modgh->errstr) {
	GEOERR("Error saving $data->{ident}: " . $modgh->errstr . "\n");
    }
}

# display the geocache in all it's glory.
sub sdisplay {
    my ($self, $dologs) = @_;
    my $data = $self->{'data'};
    my $out;
    $out .= sprintf("%-8s   %-30s Location:  %s, %s\n",
	   $data->{'ident'}, $data->{'type'} . " / " .  $data->{'subtype'},
	   sprint_coords($data->{'lat'}, $data->{'lon'}));
    $out .= sprintf("Name:      %s\n", $self->{'data'}{'groundspeak_name'});
    my $time = $data->{'time'};
    $time =~ s/T.*//;
    $out .= sprintf("Owner:     %-20s (#%d)  Created:  %s\n",
	   $data->{'groundspeak_owner_name'},
	   $data->{'groundspeak_owner_id'},
	   $time,
	  );
    $out .= sprintf("Placed By: %s\n", $data->{'groundspeak_placed_by'});
    $out .= sprintf("Status:    DISABLED\n") if ($data->{'available'} ne 'True');
    $out .= sprintf("Status:    ARCHIVED\n") if ($data->{'archived'} eq 'True');
    $out .= sprintf("Rating:    %s/%s\n",
	   $data->{'groundspeak_difficulty'}, $data->{'groundspeak_terrain'});
    $out .= sprintf("URL:       %s\n", $data->{'url'});
    $out .= sprintf("Distance:  %s\n", $data->{'distance'}) if ($data->{'distance'});
    $out .= sprintf("Short Description:\n");
    $out .= format_output($data->{'groundspeak_short_description'},
			$data->{'groundspeak_short_description_html'});

    $out .= sprintf("\nLong Description:\n");
    $out .= format_output($data->{'groundspeak_long_description'},
			$data->{'groundspeak_long_description_html'});

    $out .= $self->sshow_attributes();
    $out .= $self->sshow_tags();
    $out .= $self->sshow_ratings();
    $out .= $self->sshow_logs() if ($dologs);
    return $out;
}

sub remove_from_specific_tables {
    my $self = shift;
    $self->remove_from_a_table_by_ident('geocaches');
}

1;
